"""Defines a standard task interface for training reinforcement learning agents."""

__all__ = [
    "RLConfig",
    "RLTask",
    "RolloutConstants",
    "RolloutSharedState",
    "RolloutEnvState",
    "JointMetadataOutput",
]

import bdb
import datetime
import functools
import io
import itertools
import logging
import signal
import sys
import textwrap
import time
import traceback
from abc import ABC, abstractmethod
from collections import Counter
from dataclasses import dataclass, replace as dataclass_replace
from pathlib import Path
from threading import Thread
from typing import Any, Callable, Collection, Generic, TypeVar

import chex
import equinox as eqx
import jax
import jax.numpy as jnp
import matplotlib
import matplotlib.pyplot as plt
import mujoco
import numpy as np
import optax
import tqdm
import xax
from dpshdl.dataset import Dataset
from jaxtyping import Array, PRNGKeyArray, PyTree
from kscale.web.gen.api import JointMetadataOutput
from mujoco import mjx
from omegaconf import MISSING, DictConfig, OmegaConf
from PIL import Image, ImageDraw

from ksim.actuators import Actuators
from ksim.commands import Command
from ksim.curriculum import Curriculum, CurriculumState
from ksim.dataset import TrajectoryDataset
from ksim.engine import (
    PhysicsEngine,
    engine_type_from_physics_model,
    get_physics_engine,
)
from ksim.events import Event
from ksim.observation import Observation, ObservationInput, StatefulObservation
from ksim.randomization import PhysicsRandomizer
from ksim.resets import Reset
from ksim.rewards import Reward, StatefulReward
from ksim.terminations import Termination
from ksim.types import (
    Action,
    Histogram,
    LoggedTrajectory,
    Metrics,
    PhysicsData,
    PhysicsModel,
    PhysicsState,
    RewardState,
    Trajectory,
)
from ksim.utils.mujoco import (
    get_joint_metadata,
    get_joint_names_in_order,
    get_position_limits,
    get_torque_limits,
    load_model,
)
from ksim.viewer import DefaultMujocoViewer, GlfwMujocoViewer, RenderMode
from ksim.vis import Marker, configure_scene

logger = logging.getLogger(__name__)


@jax.tree_util.register_dataclass
@dataclass(frozen=True)
class RolloutEnvState:
    """Per-environment variables for the rollout loop."""

    commands: xax.FrozenDict[str, Array]
    physics_state: PhysicsState
    randomization_dict: xax.FrozenDict[str, Array]
    model_carry: PyTree
    reward_carry: xax.FrozenDict[str, Array]
    obs_carry: xax.FrozenDict[str, PyTree]
    curriculum_state: CurriculumState
    rng: PRNGKeyArray


@jax.tree_util.register_dataclass
@dataclass(frozen=True)
class RolloutSharedState:
    """Variables used across all environments."""

    physics_model: PhysicsModel
    model_arrs: tuple[PyTree, ...]
    aux_values: xax.FrozenDict[str, PyTree]


@jax.tree_util.register_dataclass
@dataclass(frozen=True)
class RolloutConstants:
    """Constants for the rollout loop."""

    model_statics: tuple[PyTree, ...]
    engine: PhysicsEngine
    observations: Collection[Observation]
    commands: Collection[Command]
    rewards: Collection[Reward]
    terminations: Collection[Termination]
    curriculum: Curriculum
    argmax_action: bool
    aux_constants: xax.FrozenDict[str, PyTree]


def get_observation(
    rollout_env_state: RolloutEnvState,
    observations: Collection[Observation],
    obs_carry: PyTree,
    curriculum_level: Array,
    rng: PRNGKeyArray,
) -> tuple[xax.FrozenDict[str, Array], xax.FrozenDict[str, PyTree]]:
    """Get the observation and carry from the physics state."""
    observation_dict: dict[str, Array] = {}
    next_obs_carry: dict[str, PyTree] = {}
    for observation in observations:
        rng, obs_rng, noise_rng = jax.random.split(rng, 3)
        observation_state = ObservationInput(
            commands=rollout_env_state.commands,
            physics_state=rollout_env_state.physics_state,
            obs_carry=obs_carry[observation.observation_name],
        )

        # Calls the observation function.
        if isinstance(observation, StatefulObservation):
            observation_val, new_carry = observation.observe_stateful(observation_state, curriculum_level, obs_rng)
        else:
            observation_val = observation.observe(observation_state, curriculum_level, obs_rng)
            new_carry = observation_state.obs_carry
        observation_val = observation.add_noise(observation_val, curriculum_level, noise_rng)

        observation_dict[observation.observation_name] = observation_val
        next_obs_carry[observation.observation_name] = new_carry
        rng = jax.random.split(rng)[1]

    return xax.FrozenDict(observation_dict), xax.FrozenDict(next_obs_carry)


def get_rewards(
    trajectory: Trajectory,
    rewards: Collection[Reward],
    rewards_carry: xax.FrozenDict[str, PyTree],
    rollout_length_steps: int,
    curriculum_level: Array,
    clip_min: float | None = None,
    clip_max: float | None = None,
) -> RewardState:
    """Get the rewards from the physics state."""
    reward_dict: dict[str, Array] = {}
    next_reward_carry: dict[str, PyTree] = {}
    target_shape = trajectory.done.shape

    for reward in rewards:
        reward_name = reward.reward_name
        reward_carry = rewards_carry[reward_name]

        if isinstance(reward, StatefulReward):
            reward_val, reward_carry = reward.get_reward_stateful(trajectory, reward_carry)
        else:
            reward_val = reward.get_reward(trajectory)
        reward_val = reward_val * reward.scale / rollout_length_steps
        if reward.scale_by_curriculum:
            reward_val = reward_val * curriculum_level

        if reward_val.shape != trajectory.done.shape:
            raise AssertionError(f"Reward {reward_name} shape {reward_val.shape} does not match {target_shape}")
        reward_dict[reward_name] = reward_val
        next_reward_carry[reward_name] = reward_carry

    total_reward = jax.tree.reduce(jnp.add, list(reward_dict.values()))
    if clip_min is not None:
        total_reward = jnp.maximum(total_reward, clip_min)
    if clip_max is not None:
        total_reward = jnp.minimum(total_reward, clip_max)

    return RewardState(
        total=total_reward,
        components=xax.FrozenDict(reward_dict),
        carry=xax.FrozenDict(next_reward_carry),
    )


def get_initial_obs_carry(
    rng: PRNGKeyArray,
    observations: Collection[Observation],
) -> xax.FrozenDict[str, PyTree]:
    """Get the initial observation carry."""
    rngs = jax.random.split(rng, len(observations))
    return xax.FrozenDict(
        {
            obs.observation_name: (obs.initial_carry(rng) if isinstance(obs, StatefulObservation) else None)
            for obs, rng in zip(observations, rngs, strict=True)
        }
    )


def get_initial_reward_carry(
    rng: PRNGKeyArray,
    rewards: Collection[Reward],
) -> xax.FrozenDict[str, PyTree]:
    """Get the initial reward carry."""
    rngs = jax.random.split(rng, len(rewards))
    return xax.FrozenDict(
        {
            reward.reward_name: reward.initial_carry(rng) if isinstance(reward, StatefulReward) else None
            for reward, rng in zip(rewards, rngs, strict=True)
        }
    )


def get_terminations(
    physics_state: PhysicsState,
    terminations: Collection[Termination],
    curriculum_level: Array,
) -> xax.FrozenDict[str, Array]:
    """Get the terminations from the physics state."""
    termination_dict = {}
    for termination in terminations:
        termination_val = termination(physics_state.data, curriculum_level)
        chex.assert_type(termination_val, int)
        name = termination.termination_name
        termination_dict[name] = termination_val
    return xax.FrozenDict(termination_dict)


def get_commands(
    prev_commands: xax.FrozenDict[str, Array],
    physics_state: PhysicsState,
    rng: PRNGKeyArray,
    commands: Collection[Command],
    curriculum_level: Array,
) -> xax.FrozenDict[str, Array]:
    """Get the commands from the physics state."""
    command_dict = {}
    for command_generator in commands:
        rng, cmd_rng = jax.random.split(rng)
        command_name = command_generator.command_name
        prev_command = prev_commands[command_name]
        assert isinstance(prev_command, Array)
        command_val = command_generator(prev_command, physics_state.data, curriculum_level, cmd_rng)
        command_dict[command_name] = command_val
    return xax.FrozenDict(command_dict)


def get_initial_commands(
    rng: PRNGKeyArray,
    physics_data: PhysicsData,
    commands: Collection[Command],
    curriculum_level: Array,
) -> xax.FrozenDict[str, Array]:
    """Get the initial commands from the physics state."""
    command_dict = {}
    for command_generator in commands:
        rng, cmd_rng = jax.random.split(rng)
        command_name = command_generator.command_name
        command_val = command_generator.initial_command(physics_data, curriculum_level, cmd_rng)
        command_dict[command_name] = command_val
    return xax.FrozenDict(command_dict)


def get_physics_randomizers(
    physics_model: PhysicsModel,
    randomizers: Collection[PhysicsRandomizer],
    rng: PRNGKeyArray,
) -> xax.FrozenDict[str, Array]:
    all_randomizations: dict[str, dict[str, Array]] = {}
    for randomizer in randomizers:
        rng, randomization_rng = jax.random.split(rng)
        all_randomizations[randomizer.randomization_name] = randomizer(physics_model, randomization_rng)
    for name, count in Counter([k for d in all_randomizations.values() for k in d.keys()]).items():
        if count > 1:
            name_to_keys = {k: set(v.keys()) for k, v in all_randomizations.items()}
            raise ValueError(f"Found duplicate randomization keys: {name}. Randomizations: {name_to_keys}")
    return xax.FrozenDict({k: v for d in all_randomizations.values() for k, v in d.items()})


def apply_randomizations(
    physics_model: PhysicsModel,
    engine: PhysicsEngine,
    randomizers: Collection[PhysicsRandomizer],
    curriculum_level: Array,
    rng: PRNGKeyArray,
) -> tuple[xax.FrozenDict[str, Array], PhysicsState]:
    rand_rng, reset_rng = jax.random.split(rng)
    randomizations = get_physics_randomizers(physics_model, randomizers, rand_rng)

    # Applies the randomizations to the model.
    if isinstance(physics_model, mjx.Model):
        physics_model = physics_model.tree_replace(randomizations)
    elif isinstance(physics_model, mujoco.MjModel):
        for k, v in randomizations.items():
            setattr(physics_model, k, v)
    else:
        raise ValueError(f"Unknown physics model type: {type(physics_model)}")

    physics_state = engine.reset(physics_model, curriculum_level, reset_rng)
    return randomizations, physics_state


@jax.tree_util.register_dataclass
@dataclass
class RLConfig(xax.Config):
    # Toggle this to run the environment viewer loop.
    run_model_viewer: bool = xax.field(
        value=False,
        help="Instead of dropping into the training loop, run the environment loop.",
    )
    run_model_viewer_argmax_action: bool = xax.field(
        value=True,
        help="If set, take the argmax action instead of sampling from the action distribution.",
    )
    run_viewer_num_seconds: float | None = xax.field(
        value=None,
        help="If provided, run the environment loop for the given number of seconds.",
    )
    run_viewer_save_renders: bool = xax.field(
        value=False,
        help="If set, save the renders to the experiment directory.",
    )
    run_viewer_save_video: bool = xax.field(
        value=False,
        help="If set, render the environment as a video instead of a GIF.",
    )

    # Toggle this to collect a dataset.
    collect_dataset: bool = xax.field(
        value=False,
        help="If true, collect a dataset.",
    )
    dataset_num_batches: int = xax.field(
        1,
        help="The number of trajectories to collect at a time.",
    )
    dataset_save_path: str | None = xax.field(
        value=None,
        help="If provided, save the dataset to the given path.",
    )
    collect_dataset_argmax_action: bool = xax.field(
        value=False,
        help="If set, get the argmax action, otherwise sample randomly from the model.",
    )

    # Logging parameters.
    log_train_metrics: bool = xax.field(
        value=True,
        help="If true, log train metrics.",
    )
    epochs_per_log_step: int = xax.field(
        value=1,
        help="The number of epochs between logging steps.",
    )
    profile_memory: bool = xax.field(
        value=False,
        help="If true, profile memory usage.",
    )

    # Training parameters.
    num_envs: int = xax.field(
        value=MISSING,
        help="The number of training environments to run in parallel.",
    )
    batch_size: int = xax.field(
        value=1,
        help="The number of model update batches per trajectory batch. ",
    )
    rollout_length_seconds: float = xax.field(
        value=MISSING,
        help="The number of seconds to rollout each environment during training.",
    )

    # Validation timing parameters.
    valid_every_n_seconds: float | None = xax.field(
        value=150.0,
        help="Run full validation (render trajectory and all graphs) every N seconds",
    )
    valid_first_n_seconds: float | None = xax.field(
        value=None,
        help="Run first validation after N seconds",
    )
    render_full_every_n_steps: int = xax.field(
        value=12,
        help="Render the trajectory (without associated graphs) every N seconds",
    )

    # Rendering parameters.
    max_values_per_plot: int = xax.field(
        value=8,
        help="The maximum number of values to plot for each key.",
    )
    plot_figsize: tuple[float, float] = xax.field(
        value=(8, 4),
        help="The size of the figure for each plot.",
    )
    render_with_glfw: bool | None = xax.field(
        value=None,
        help="Explicitly toggle GLFW rendering; if not specified, use GLFW when rendering on-screen",
    )
    render_shadow: bool = xax.field(
        value=False,
        help="If true, render shadows.",
    )
    render_reflection: bool = xax.field(
        value=False,
        help="If true, render reflections.",
    )
    render_contact_force: bool = xax.field(
        value=False,
        help="If true, render contact forces.",
    )
    render_contact_point: bool = xax.field(
        value=False,
        help="If true, render contact points.",
    )
    render_inertia: bool = xax.field(
        value=False,
        help="If true, render inertia.",
    )
    render_height_small: int = xax.field(
        value=240,
        help="The height of the rendered images.",
    )
    render_width_small: int = xax.field(
        value=360,
        help="The width of the rendered images.",
    )
    render_height: int = xax.field(
        value=480,
        help="The height of the rendered images during the validation phase.",
    )
    render_width: int = xax.field(
        value=640,
        help="The height of the rendered images during the validation phase.",
    )
    render_length_seconds: float | None = xax.field(
        value=5.0,
        help="The number of seconds to rollout each environment during evaluation.",
    )
    render_fps: int | None = xax.field(
        value=24,
        help="The target FPS for the renderered video.",
    )
    render_slowdown: float = xax.field(
        value=2.0,
        help="The slowdown factor for the rendered video.",
    )
    render_track_body_id: int | None = xax.field(
        value=None,
        help="If set, the render camera will track the body with this ID.",
    )
    render_distance: float = xax.field(
        value=3.5,
        help="The distance of the camera from the target.",
    )
    render_azimuth: float = xax.field(
        value=90.0,
        help="The azimuth of the render camera.",
    )
    render_elevation: float = xax.field(
        value=-10.0,
        help="The elevation of the render camera.",
    )
    render_lookat: list[float] = xax.field(
        value=[0.0, 0.0, 0.5],
        help="The lookat point of the render camera.",
    )

    # Engine parameters.
    ctrl_dt: float = xax.field(
        value=0.02,
        help="The time step of the control loop.",
    )
    dt: float = xax.field(
        value=0.002,
        help="The time step of the physics loop.",
    )
    tolerance: float = xax.field(
        value=1e-10,
        help="The tolerance of the solver.",
    )
    iterations: int = xax.field(
        value=MISSING,
        help="Number of main solver iterations",
    )
    ls_iterations: int = xax.field(
        value=MISSING,
        help="Maximum number of CG / Newton linesearch iterations",
    )
    solver: str = xax.field(
        value="newton",
        help="The constraint solver algorithm to use",
    )
    integrator: str = xax.field(
        value="implicitfast",
        help="The integrator algorithm to use",
    )
    disable_euler_damping: bool = xax.field(
        value=True,
        help="If set, disable Euler damping - this is a performance improvement",
    )
    max_action_latency: float = xax.field(
        value=0.0,
        help="The maximum latency of the action.",
    )
    reward_clip_min: float | None = xax.field(
        value=None,
        help="The minimum value of the reward.",
    )
    reward_clip_max: float | None = xax.field(
        value=None,
        help="The maximum value of the reward.",
    )
    render_markers: bool = xax.field(
        value=False,
        help="If true, render markers.",
    )
    render_camera_name: str | int | None = xax.field(
        value=None,
        help="The name or id of the camera to use in rendering.",
    )


Config = TypeVar("Config", bound=RLConfig)


def get_viewer(
    mj_model: mujoco.MjModel,
    config: Config,
    mj_data: mujoco.MjData | None = None,
    save_path: str | Path | None = None,
    mode: RenderMode | None = None,
    is_small: bool = False,
) -> GlfwMujocoViewer | DefaultMujocoViewer:
    if mode is None:
        mode = "window" if save_path is None else "offscreen"

    if (render_with_glfw := config.render_with_glfw) is None:
        render_with_glfw = mode == "window"

    viewer: GlfwMujocoViewer | DefaultMujocoViewer

    if render_with_glfw:
        viewer = GlfwMujocoViewer(
            mj_model,
            data=mj_data,
            mode=mode,
            width=config.render_width_small if is_small else config.render_width,
            height=config.render_height_small if is_small else config.render_height,
            shadow=config.render_shadow,
            reflection=config.render_reflection,
            contact_force=config.render_contact_force,
            contact_point=config.render_contact_point,
            inertia=config.render_inertia,
        )

    else:
        viewer = DefaultMujocoViewer(
            mj_model,
            width=config.render_width_small if is_small else config.render_width,
            height=config.render_height_small if is_small else config.render_height,
        )

    # Sets the viewer camera.
    viewer.cam.distance = config.render_distance
    viewer.cam.azimuth = config.render_azimuth
    viewer.cam.elevation = config.render_elevation
    viewer.cam.lookat[:] = config.render_lookat
    if config.render_track_body_id is not None:
        viewer.cam.trackbodyid = config.render_track_body_id
        viewer.cam.type = mujoco.mjtCamera.mjCAMERA_TRACKING

    if config.render_camera_name is not None:
        viewer.set_camera(config.render_camera_name)

    configure_scene(
        viewer.scn,
        viewer.vopt,
        shadow=config.render_shadow,
        contact_force=config.render_contact_force,
        contact_point=config.render_contact_point,
        inertia=config.render_inertia,
    )

    return viewer


@jax.tree_util.register_dataclass
@dataclass(frozen=True)
class RLLoopConstants:
    optimizer: tuple[optax.GradientTransformation, ...]
    constants: RolloutConstants


@jax.tree_util.register_dataclass
@dataclass(frozen=True)
class RLLoopCarry:
    opt_state: tuple[optax.OptState, ...]
    env_states: RolloutEnvState
    shared_state: RolloutSharedState


class RLTask(xax.Task[Config], Generic[Config], ABC):
    """Base class for reinforcement learning tasks."""

    def __init__(self, config: Config) -> None:
        super().__init__(config)

        # Using this Matplotlib backend since it is non-interactive.
        matplotlib.use("agg")

        if self.config.num_envs % self.config.batch_size != 0:
            raise ValueError(
                f"The number of environments ({self.config.num_envs}) must be divisible by "
                f"the batch size ({self.config.batch_size})"
            )

    @functools.cached_property
    def batch_size(self) -> int:
        return self.config.batch_size

    @functools.cached_property
    def num_batches(self) -> int:
        return self.config.num_envs // self.batch_size

    @abstractmethod
    def get_mujoco_model(self) -> mujoco.MjModel: ...

    def set_mujoco_model_opts(self, mj_model: mujoco.MjModel) -> mujoco.MjModel:
        def _set_opt(name: str, value: Any) -> None:  # noqa: ANN401
            model_val = getattr(mj_model.opt, name)
            if model_val != value:
                logger.debug("User-specified %s %s is different from model %s %s", name, value, name, model_val)
                setattr(mj_model.opt, name, value)

        solver = getattr(mjx.SolverType, self.config.solver.upper(), None)
        if solver is None:
            raise ValueError(f"Invalid solver type: {self.config.solver}")

        integrator = getattr(mjx.IntegratorType, self.config.integrator.upper(), None)
        if integrator is None:
            raise ValueError(f"Invalid integrator type: {self.config.integrator}")

        _set_opt("timestep", self.config.dt)
        _set_opt("iterations", self.config.iterations)
        _set_opt("ls_iterations", self.config.ls_iterations)
        _set_opt("integrator", integrator)
        _set_opt("tolerance", self.config.tolerance)
        _set_opt("solver", solver)

        if self.config.disable_euler_damping:
            mj_model.opt.disableflags = mj_model.opt.disableflags | mjx.DisableBit.EULERDAMP

        return mj_model

    def get_mujoco_model_metadata(self, mj_model: mujoco.MjModel) -> dict[str, JointMetadataOutput]:
        return get_joint_metadata(mj_model)

    def get_mjx_model(self, mj_model: mujoco.MjModel) -> mjx.Model:
        """Convert a mujoco model to an mjx model.

        Args:
            mj_model: The mujoco model to convert.

        Returns:
            The mjx model.
        """
        return load_model(mj_model)

    def get_engine(
        self,
        physics_model: PhysicsModel,
        metadata: dict[str, JointMetadataOutput] | None = None,
    ) -> PhysicsEngine:
        return get_physics_engine(
            engine_type=engine_type_from_physics_model(physics_model),
            resets=self.get_resets(physics_model),
            events=self.get_events(physics_model),
            actuators=self.get_actuators(physics_model, metadata),
            dt=float(physics_model.opt.timestep),
            ctrl_dt=self.config.ctrl_dt,
            max_action_latency=self.config.max_action_latency,
        )

    @abstractmethod
    def get_physics_randomizers(self, physics_model: PhysicsModel) -> Collection[PhysicsRandomizer]:
        """Returns randomizers, for randomizing each environment.

        Args:
            physics_model: The physics model to get the randomization for.

        Returns:
            A collection of randomization generators.
        """

    @abstractmethod
    def get_resets(self, physics_model: PhysicsModel) -> Collection[Reset]:
        """Returns the reset generators for the current task.

        Args:
            physics_model: The physics model to get the resets for.

        Returns:
            A collection of reset generators.
        """

    @abstractmethod
    def get_events(self, physics_model: PhysicsModel) -> Collection[Event]:
        """Returns the event generators for the current task.

        Args:
            physics_model: The physics model to get the events for.
        """

    @abstractmethod
    def get_actuators(
        self,
        physics_model: PhysicsModel,
        metadata: dict[str, JointMetadataOutput] | None = None,
    ) -> Actuators: ...

    @abstractmethod
    def get_observations(self, physics_model: PhysicsModel) -> Collection[Observation]:
        """Returns the observation generators for the current task.

        Args:
            physics_model: The physics model to get the observations for.

        Returns:
            A collection of observation generators.
        """

    @abstractmethod
    def get_commands(self, physics_model: PhysicsModel) -> Collection[Command]:
        """Returns the command generators for the current task.

        Args:
            physics_model: The physics model to get the commands for.

        Returns:
            A collection of command generators.
        """

    @abstractmethod
    def get_rewards(self, physics_model: PhysicsModel) -> Collection[Reward]:
        """Returns the reward generators for the current task.

        Args:
            physics_model: The physics model to get the rewards for.

        Returns:
            A collection of reward generators.
        """

    @abstractmethod
    def get_terminations(self, physics_model: PhysicsModel) -> Collection[Termination]:
        """Returns the termination generators for the current task.

        Args:
            physics_model: The physics model to get the terminations for.

        Returns:
            A collection of termination generators.
        """

    @abstractmethod
    def get_initial_model_carry(self, rng: PRNGKeyArray) -> PyTree | None:
        """Returns the initial carry for the model.

        Args:
            rng: The random key to use.

        Returns:
            An arbitrary PyTree, representing any carry parameters that the
            model needs.
        """

    @abstractmethod
    def get_curriculum(self, physics_model: PhysicsModel) -> Curriculum:
        """Returns the curriculum for the current task.

        Args:
            physics_model: The physics model to get the curriculum for.
        """

    @abstractmethod
    def sample_action(
        self,
        model: PyTree,
        model_carry: PyTree,
        physics_model: PhysicsModel,
        physics_state: PhysicsState,
        observations: xax.FrozenDict[str, Array],
        commands: xax.FrozenDict[str, Array],
        rng: PRNGKeyArray,
        argmax: bool,
    ) -> Action:
        """Gets an action for the current observation.

        This function returns the action to take, the next carry (for models
        which look at multiple steps), and any auxiliary outputs. The auxiliary
        outputs get stored in the final trajectory object and can be used to
        compute metrics like log probabilities, values, etc.

        Args:
            model: The current model.
            physics_model: The physics model.
            physics_state: The current physics state.
            observations: The current observations.
            commands: The current commands.
            model_carry: The model carry from the previous step.
            rng: The random key.
            argmax: If set, get the argmax action, otherwise sample randomly
                from the model.

        Returns:
            The action to take, the next carry, and any auxiliary outputs.
        """

    @property
    def rollout_length_steps(self) -> int:
        return round(self.config.rollout_length_seconds / self.config.ctrl_dt)

    @property
    def rollout_num_samples(self) -> int:
        return self.rollout_length_steps * self.config.num_envs

    def get_mujoco_model_info(self, mj_model: mujoco.MjModel) -> dict:
        return {
            "joint_names": get_joint_names_in_order(mj_model),
            "position_limits": get_position_limits(mj_model),
            "torque_limits": get_torque_limits(mj_model),
        }

    @xax.jit(static_argnames=["self", "constants"], jit_level=3)
    def step_engine(
        self,
        constants: RolloutConstants,
        env_states: RolloutEnvState,
        shared_state: RolloutSharedState,
    ) -> tuple[Trajectory, RolloutEnvState]:
        """Runs a single step of the physics engine.

        Args:
            constants: The constants for the engine.
            env_states: The environment variables for the engine.
            shared_state: The control variables for the engine.

        Returns:
            A tuple containing the trajectory and the next engine variables.
        """
        rng = env_states.rng
        rng, obs_rng, cmd_rng, act_rng, reset_rng, carry_rng, physics_rng = jax.random.split(rng, 7)

        # Recombines the mutable and static parts of the model.
        policy_model_arr = shared_state.model_arrs[0]
        policy_model_static = constants.model_statics[0]
        policy_model = eqx.combine(policy_model_arr, policy_model_static)

        # Gets the observations from the physics state.
        observations, next_obs_carry = get_observation(
            rollout_env_state=env_states,
            observations=constants.observations,
            obs_carry=env_states.obs_carry,
            curriculum_level=env_states.curriculum_state.level,
            rng=obs_rng,
        )

        # Samples an action from the model.
        action = self.sample_action(
            model=policy_model,
            model_carry=env_states.model_carry,
            physics_model=shared_state.physics_model,
            physics_state=env_states.physics_state,
            observations=observations,
            commands=env_states.commands,
            rng=act_rng,
            argmax=constants.argmax_action,
        )

        # Steps the physics engine.
        next_physics_state: PhysicsState = constants.engine.step(
            action=action.action,
            physics_model=shared_state.physics_model,
            physics_state=env_states.physics_state,
            curriculum_level=env_states.curriculum_state.level,
            rng=physics_rng,
        )

        # Gets termination components and a single termination boolean.
        terminations = get_terminations(
            physics_state=next_physics_state,
            terminations=constants.terminations,
            curriculum_level=env_states.curriculum_state.level,
        )

        # Convert ternary terminations to binary arrays.
        terminated = jax.tree.reduce(jnp.logical_or, [t != 0 for t in terminations.values()])
        success = jax.tree.reduce(jnp.logical_and, [t != -1 for t in terminations.values()]) & terminated

        # Combines all the relevant data into a single object. Lives up here to
        # avoid accidentally incorporating information it shouldn't access to.
        transition = Trajectory(
            qpos=jnp.array(next_physics_state.data.qpos),
            qvel=jnp.array(next_physics_state.data.qvel),
            xpos=jnp.array(next_physics_state.data.xpos),
            xquat=jnp.array(next_physics_state.data.xquat),
            obs=observations,
            command=env_states.commands,
            event_state=next_physics_state.event_states,
            action=action.action,
            done=terminated,
            success=success,
            timestep=next_physics_state.data.time,
            termination_components=terminations,
            aux_outputs=action.aux_outputs,
        )

        # Conditionally reset on termination.
        next_commands = jax.lax.cond(
            terminated,
            lambda: get_initial_commands(
                rng=cmd_rng,
                physics_data=next_physics_state.data,
                commands=constants.commands,
                curriculum_level=env_states.curriculum_state.level,
            ),
            lambda: get_commands(
                prev_commands=env_states.commands,
                physics_state=next_physics_state,
                rng=cmd_rng,
                commands=constants.commands,
                curriculum_level=env_states.curriculum_state.level,
            ),
        )

        next_obs_carry = jax.lax.cond(
            terminated,
            lambda: get_initial_obs_carry(rng=carry_rng, observations=constants.observations),
            lambda: next_obs_carry,
        )

        next_physics_state = jax.lax.cond(
            terminated,
            lambda: constants.engine.reset(
                shared_state.physics_model,
                env_states.curriculum_state.level,
                reset_rng,
            ),
            lambda: next_physics_state,
        )

        next_model_carry = jax.lax.cond(
            terminated,
            lambda: self.get_initial_model_carry(carry_rng),
            lambda: action.carry,
        )

        # Gets the variables for the next step.
        next_env_state = dataclass_replace(
            env_states,
            commands=next_commands,
            physics_state=next_physics_state,
            model_carry=next_model_carry,
            obs_carry=next_obs_carry,
            rng=rng,
        )

        return transition, next_env_state

    def get_dataset(self, phase: xax.Phase) -> Dataset:
        raise NotImplementedError("RL tasks do not require datasets, since trajectory histories are stored in-memory.")

    def compute_loss(self, model: PyTree, batch: Any, output: Any, state: xax.State) -> Array:  # noqa: ANN401
        raise NotImplementedError(
            "Direct compute_loss from TrainMixin is not expected to be called in RL tasks. "
            "PPO tasks use model_update and loss_metrics_grads instead."
        )

    def run(self) -> None:
        """Highest level entry point for RL tasks, determines what to run."""
        if self.config.run_model_viewer:
            self.run_model_viewer(
                num_steps=(
                    None
                    if self.config.run_viewer_num_seconds is None
                    else round(self.config.run_viewer_num_seconds / self.config.ctrl_dt)
                ),
                save_renders=self.config.run_viewer_save_renders,
                argmax_action=self.config.run_model_viewer_argmax_action,
            )

        elif self.config.collect_dataset:
            self.collect_dataset(
                num_batches=self.config.dataset_num_batches,
                save_path=self.config.dataset_save_path,
                argmax_action=self.config.collect_dataset_argmax_action,
            )

        else:
            self.run_training()

    def log_train_metrics(self, metrics: Metrics) -> None:
        """Logs the train metrics.

        Args:
            metrics: The metrics to log.
            rollout_length: The length of the rollout.
        """
        if self.config.log_train_metrics:
            for namespace, metric, secondary in (
                ("🚂 train", metrics.train, True),
                ("🎁 reward", metrics.reward, False),
                ("💀 termination", metrics.termination, True),
                ("🔄 curriculum", {"level": metrics.curriculum_level}, True),
            ):
                for key, value in metric.items():
                    if isinstance(value, Histogram):
                        self.logger.log_histogram_raw(
                            key,
                            counts=value.counts,
                            limits=value.limits,
                            minv=value.min,
                            maxv=value.max,
                            sumv=value.sum,
                            sum_squaresv=value.sum_squares,
                            namespace=f"{namespace} histograms",
                        )
                        self.logger.log_scalar(key, value.mean, namespace=namespace, secondary=secondary)
                    else:
                        self.logger.log_scalar(key, value.mean(), namespace=namespace, secondary=secondary)

    def render_trajectory_video(
        self,
        trajectory: Trajectory,
        markers: Collection[Marker],
        viewer: GlfwMujocoViewer | DefaultMujocoViewer,
        target_fps: int | None = None,
    ) -> tuple[np.ndarray, int]:
        """Render trajectory as video frames with computed FPS."""
        fps = round(1 / self.config.ctrl_dt)

        # Rerenders the trajectory at the desired FPS.
        num_frames = len(trajectory.done)
        if target_fps is not None:
            indices = jnp.arange(0, num_frames, fps / target_fps, dtype=jnp.int32).clip(max=num_frames - 1)
            trajectory = jax.tree.map(lambda arr: arr[indices], trajectory)
            fps = target_fps
        else:
            indices = jnp.arange(0, num_frames, dtype=jnp.int32)

        chex.assert_shape(trajectory.done, (None,))
        num_steps = trajectory.done.shape[0]

        trajectory_list: list[Trajectory] = [
            jax.tree.map(lambda arr, i=i: arr[i], trajectory) for i in range(num_steps)
        ]

        frame_list: list[np.ndarray] = []

        for frame_id, sub_trajectory in enumerate(trajectory_list):
            # Updates the model with the latest data.
            viewer.data.qpos[:] = np.array(sub_trajectory.qpos)
            viewer.data.qvel[:] = np.array(sub_trajectory.qvel)
            mujoco.mj_forward(viewer.model, viewer.data)

            def render_callback(
                model: mujoco.MjModel,
                data: mujoco.MjData,
                scene: mujoco.MjvScene,
                traj: Trajectory = sub_trajectory,
            ) -> None:
                if self.config.render_markers:
                    for marker in markers:
                        marker(model, data, scene, traj)

            frame = viewer.read_pixels(callback=render_callback)

            # Overlays the frame number on the frame.
            frame_img = Image.fromarray(frame)
            draw = ImageDraw.Draw(frame_img)
            draw.text((10, 10), f"Frame {indices[frame_id]}", fill=(255, 255, 255))
            frame = np.array(frame_img)

            # Draws an RGB patch in the bottom right corner of the frame.
            rgb = np.zeros((5, 25, 3), dtype=np.uint8)
            rgb[:, 5:10, 0] = 255
            rgb[:, 10:15, 1] = 255
            rgb[:, 15:20, 2] = 255
            rgb[:, 20:, :] = 255
            frame[-5:, -25:] = rgb

            frame_list.append(frame)

        return np.stack(frame_list, axis=0), fps

    def _crop_to_length(self, logged_traj: LoggedTrajectory, length: float) -> LoggedTrajectory:
        render_frames = round(length / self.config.ctrl_dt)
        return jax.tree.map(lambda arr: arr[:render_frames], logged_traj)

    def _log_logged_trajectory_graphs(
        self,
        logged_traj: LoggedTrajectory,
        log_callback: Callable[[str, Image.Image, str], None],
    ) -> None:
        """Visualizes a single trajectory.

        Args:
            logged_traj: The single trajectory to log.
            log_callback: A callable function to run to log a given image.
        """
        # Clips the trajectory to the desired length.
        if self.config.render_length_seconds is not None:
            logged_traj = self._crop_to_length(logged_traj, self.config.render_length_seconds)

        # Logs plots of the observations, commands, actions, rewards, and terminations.
        # Emojis are used in order to prevent conflicts with user-specified namespaces.
        for namespace, arr_dict in (
            ("👀 obs images", logged_traj.trajectory.obs),
            ("🕹️ command images", logged_traj.trajectory.command),
            ("🏃 action images", {"action": logged_traj.trajectory.action}),
            ("💀 termination images", logged_traj.trajectory.termination_components),
            ("🗓️ event images", logged_traj.trajectory.event_state),
            ("🎁 reward images", logged_traj.rewards.components),
            ("🎁 reward images", {"total": logged_traj.rewards.total}),
            ("📈 metrics images", logged_traj.metrics),
        ):
            for key, value in arr_dict.items():
                plt.figure(figsize=self.config.plot_figsize)

                # Ensures a consistent shape and truncates if necessary.
                value = value.reshape(value.shape[0], -1)
                if value.shape[-1] > self.config.max_values_per_plot:
                    logger.debug("Truncating %s to %d values per plot.", key, self.config.max_values_per_plot)
                    value = value[..., : self.config.max_values_per_plot]

                for i in range(value.shape[1]):
                    plt.plot(value[:, i], label=f"{i}")

                if value.shape[1] > 1:
                    plt.legend()
                plt.title(key)

                # Converts to PIL image.
                buf = io.BytesIO()
                plt.savefig(buf, format="png")
                plt.close()
                buf.seek(0)
                img = Image.open(buf)

                # Logs the image.
                # self.logger.log_image(key=key, value=img, namespace=namespace)
                log_callback(key, img, namespace)

    def _log_logged_trajectory_video(
        self,
        logged_traj: LoggedTrajectory,
        markers: Collection[Marker],
        viewer: GlfwMujocoViewer | DefaultMujocoViewer,
        key: str,
    ) -> None:
        """Visualizes a single trajectory.

        Args:
            logged_traj: The single trajectory to log.
            markers: The markers to visualize.
            viewer: The Mujoco viewer to render the scene with.
            key: The logging key to use.
        """
        # Clips the trajectory to the desired length.
        if self.config.render_length_seconds is not None:
            logged_traj = self._crop_to_length(logged_traj, self.config.render_length_seconds)

        # Logs the video of the trajectory.
        frames, fps = self.render_trajectory_video(
            trajectory=logged_traj.trajectory,
            markers=markers,
            viewer=viewer,
            target_fps=self.config.render_fps,
        )

        self.logger.log_video(
            key=key,
            value=frames,
            fps=round(fps / self.config.render_slowdown),
            namespace="➡️ trajectory images",
        )

    @abstractmethod
    def update_model(
        self,
        *,
        constants: RLLoopConstants,
        carry: RLLoopCarry,
        trajectories: Trajectory,
        rewards: RewardState,
        rng: PRNGKeyArray,
    ) -> tuple[
        RLLoopCarry,
        xax.FrozenDict[str, Array],
        LoggedTrajectory,
    ]:
        """Updates the model on the given trajectory.

        This function should be implemented according to the specific RL method
        that we are using.

        Args:
            constants: The constants to use for the rollout.
            carry: The carry to use for the rollout.
            trajectories: The trajectories to update the model on.
            rewards: The rewards to update the model on.
            rng: The random seed.

        Returns:
            A tuple containing the next carry, metrics to log, and the single
            trajectory to log. If a metric has a single element it is logged as
            a scalar, otherwise it is logged as a histogram.
        """

    def get_histogram(self, arr: Array, bins: int = 100) -> Histogram:
        """Gets the histogram of the given array.

        Args:
            arr: The array to get the histogram for.
            bins: The number of bins to use for the histogram.

        Returns:
            The histogram of the given array.
        """
        arr = arr.flatten()
        counts, limits = jnp.histogram(arr, bins=bins)
        return Histogram(
            counts=counts,
            limits=limits[..., 1:],
            min=arr.min(),
            max=arr.max(),
            sum=arr.sum(),
            sum_squares=arr.dot(arr),
            mean=arr.mean(),
        )

    def get_reward_metrics(self, trajectories: Trajectory, rewards: RewardState) -> dict[str, Array]:
        """Gets the reward metrics.

        Args:
            trajectories: The trajectories to get the reward metrics for.
            rewards: The rewards to get the metrics for.
        """
        return {
            "total": rewards.total,
            **{key: value for key, value in rewards.components.items()},
        }

    def get_termination_metrics(self, trajectories: Trajectory) -> dict[str, Array]:
        """Gets the termination metrics.

        Args:
            trajectories: The trajectories to get the termination metrics for.
        """
        # Compute the mean number of terminations per episode, broken down by
        # the type of termination.
        kvs = list(trajectories.termination_components.items())
        all_terminations = jnp.stack([v for _, v in kvs], axis=-1)
        has_termination = (all_terminations.any(axis=-1)).sum(axis=-1)
        num_terminations = has_termination.sum().clip(min=1)
        mean_terminations = trajectories.done.sum(-1).mean()

        return {
            "episode_length": trajectories.episode_length(),
            "mean_terminations": mean_terminations,
            **{f"prct/{key}": ((value != 0).sum() / num_terminations) for key, value in kvs},
        }

    def get_markers(
        self,
        commands: Collection[Command],
        observations: Collection[Observation],
        rewards: Collection[Reward],
    ) -> Collection[Marker]:
        markers: list[Marker] = []
        for command in commands:
            markers.extend(command.get_markers())
        for observation in observations:
            markers.extend(observation.get_markers())
        for reward in rewards:
            markers.extend(reward.get_markers())
        return markers

    def postprocess_trajectory(
        self,
        constants: RolloutConstants,
        env_states: RolloutEnvState,
        shared_state: RolloutSharedState,
        trajectory: Trajectory,
    ) -> Trajectory:
        return trajectory

    @xax.jit(static_argnames=["self", "constants"], jit_level=2)
    def _single_unroll(
        self,
        constants: RolloutConstants,
        env_state: RolloutEnvState,
        shared_state: RolloutSharedState,
    ) -> tuple[Trajectory, RewardState, RolloutEnvState]:
        # Applies randomizations to the model.
        shared_state = dataclass_replace(
            shared_state,
            physics_model=shared_state.physics_model.tree_replace(env_state.randomization_dict),
        )

        def scan_fn(env_state: RolloutEnvState, _: None) -> tuple[RolloutEnvState, Trajectory]:
            trajectory, env_state = self.step_engine(
                constants=constants,
                env_states=env_state,
                shared_state=shared_state,
            )
            return env_state, trajectory

        # Scans the engine for the desired number of steps.
        env_state, trajectory = xax.scan(
            scan_fn,
            env_state,
            length=self.rollout_length_steps,
        )

        # Post-processes the trajectory.
        trajectory = self.postprocess_trajectory(
            constants=constants,
            env_states=env_state,
            shared_state=shared_state,
            trajectory=trajectory,
        )

        # Gets the rewards.
        reward = get_rewards(
            trajectory=trajectory,
            rewards=constants.rewards,
            rewards_carry=env_state.reward_carry,
            rollout_length_steps=self.rollout_length_steps,
            curriculum_level=env_state.curriculum_state.level,
            clip_min=self.config.reward_clip_min,
            clip_max=self.config.reward_clip_max,
        )

        # Updates the reward carry in the environment state.
        env_state = dataclass_replace(
            env_state,
            reward_carry=reward.carry,
        )

        return trajectory, reward, env_state

    @xax.jit(static_argnames=["self", "constants"], jit_level=1)
    def _rl_train_loop_step(
        self,
        carry: RLLoopCarry,
        constants: RLLoopConstants,
        state: xax.State,
        rng: PRNGKeyArray,
    ) -> tuple[RLLoopCarry, Metrics, LoggedTrajectory]:
        """Runs a single step of the RL training loop."""

        def single_step_fn(
            carry_i: RLLoopCarry,
            rng: PRNGKeyArray,
        ) -> tuple[RLLoopCarry, tuple[Metrics, LoggedTrajectory]]:
            # Rolls out a new trajectory.
            vmapped_unroll = jax.vmap(self._single_unroll, in_axes=(None, 0, None))
            trajectories, rewards, env_state = vmapped_unroll(
                constants.constants,
                carry_i.env_states,
                carry_i.shared_state,
            )

            # Runs update on the previous trajectory.
            carry_i, train_metrics, logged_traj = self.update_model(
                constants=constants,
                carry=carry_i,
                trajectories=trajectories,
                rewards=rewards,
                rng=rng,
            )

            # Store all the metrics to log.
            metrics = Metrics(
                train=train_metrics,
                reward=xax.FrozenDict(self.get_reward_metrics(trajectories, rewards)),
                termination=xax.FrozenDict(self.get_termination_metrics(trajectories)),
                curriculum_level=carry_i.env_states.curriculum_state.level,
            )

            # Steps the curriculum.
            curriculum_state = constants.constants.curriculum(
                trajectory=trajectories,
                rewards=rewards,
                training_state=state,
                prev_state=carry_i.env_states.curriculum_state,
            )

            # Update the environment states *after* doing the model update -
            # the model needs to be updated using the same environment states
            # that were used to generate the trajectory.
            carry_i = dataclass_replace(
                carry_i,
                env_states=dataclass_replace(
                    env_state,
                    curriculum_state=curriculum_state,
                ),
            )

            return carry_i, (metrics, logged_traj)

        rngs = jax.random.split(rng, self.config.epochs_per_log_step)
        carry, (metrics, logged_traj) = jax.lax.scan(single_step_fn, carry, rngs)

        # Convert any array with more than one element to a histogram.
        metrics = jax.tree.map(lambda x: self.get_histogram(x) if isinstance(x, Array) and x.size > 1 else x, metrics)

        # Only get final trajectory and rewards.
        logged_traj = jax.tree.map(lambda arr: arr[-1], logged_traj)

        return carry, metrics, logged_traj

    def run_environment_step(
        self,
        constants: RolloutConstants,
        env_states: RolloutEnvState,
        shared_state: RolloutSharedState,
    ) -> tuple[Array, Array, tuple[Trajectory, RolloutEnvState] | None]:
        """Runs a single step of the environment.

        Args:
            constants: The constants
            env_states: The environment states
            shared_state: The shared state

        Returns:
            A tuple containing the qpos, qvel, and optionally the
            transition and env_states.
        """
        transition, env_states = self.step_engine(
            constants=constants,
            env_states=env_states,
            shared_state=shared_state,
        )

        qpos = env_states.physics_state.data.qpos
        qvel = env_states.physics_state.data.qvel

        return qpos, qvel, (transition, env_states)

    def run_model_viewer(
        self,
        num_steps: int | None = None,
        save_renders: bool = False,
        argmax_action: bool = True,
    ) -> None:
        """Provides an easy-to-use interface for debugging environments.

        This function runs the environment for `num_steps`, rendering using
        MujocoViewer while simultaneously plotting the reward and termination
        information.

        Args:
            num_steps: The number of steps to run the environment for. If not
                provided, run until the user manually terminates the
                environment visualizer.
            save_renders: If provided, save the rendered video to the given path.
            argmax_action: If set, get the argmax action, otherwise sample
                randomly from the model.
        """
        save_path = self.exp_dir / "renders" / f"render_{time.time()}" if save_renders else None
        if save_path is not None:
            save_path.mkdir(parents=True, exist_ok=True)

        with self, jax.disable_jit():
            rng = self.prng_key()
            self.set_loggers()

            # Loads the Mujoco model and logs some information about it.
            mj_model = self.get_mujoco_model()
            mj_model = self.set_mujoco_model_opts(mj_model)
            mujoco_info = OmegaConf.to_yaml(DictConfig(self.get_mujoco_model_info(mj_model)))
            self.logger.log_file("mujoco_info.yaml", mujoco_info)

            randomizers = self.get_physics_randomizers(mj_model)

            rng, model_rng = jax.random.split(rng)
            models, _ = self.load_initial_state(model_rng, load_optimizer=False)

            # Partitions the models into mutable and static parts.
            model_arrs, model_statics = (
                tuple(models)
                for models in zip(
                    *(eqx.partition(model, self.model_partition_fn) for model in models),
                    strict=True,
                )
            )

            constants = self._get_constants(
                mj_model=mj_model,
                physics_model=mj_model,
                model_statics=model_statics,
                argmax_action=argmax_action,
            )
            env_states = self._get_env_state(
                rng=rng,
                rollout_constants=constants,
                mj_model=mj_model,
                physics_model=mj_model,
                randomizers=randomizers,
            )
            shared_state = self._get_shared_state(
                mj_model=mj_model,
                physics_model=mj_model,
                model_arrs=model_arrs,
            )

            # Creates the markers.
            markers = self.get_markers(
                commands=constants.commands,
                observations=constants.observations,
                rewards=constants.rewards,
            )

            # Creates the viewer.
            viewer = get_viewer(
                mj_model=mj_model,
                config=self.config,
                mj_data=env_states.physics_state.data,
                save_path=save_path,
            )

            iterator = tqdm.trange(num_steps) if num_steps is not None else tqdm.tqdm(itertools.count())
            frames: list[np.ndarray] = []

            transitions = []

            try:
                for _ in iterator:
                    transition, env_states = self.step_engine(
                        constants=constants,
                        env_states=env_states,
                        shared_state=shared_state,
                    )
                    transitions.append(transition)

                    # Logs the frames to render.
                    viewer.data.qpos[:] = np.array(env_states.physics_state.data.qpos)
                    viewer.data.qvel[:] = np.array(env_states.physics_state.data.qvel)
                    mujoco.mj_forward(viewer.model, viewer.data)

                    def render_callback(
                        model: mujoco.MjModel,
                        data: mujoco.MjData,
                        scene: mujoco.MjvScene,
                        traj: Trajectory = transition,
                    ) -> None:
                        if self.config.render_markers:
                            for marker in markers:
                                marker(model, data, scene, traj)

                    if save_path is None:
                        viewer.render(callback=render_callback)
                    else:
                        frames.append(viewer.read_pixels(callback=render_callback))

            except (KeyboardInterrupt, bdb.BdbQuit):
                logger.info("Keyboard interrupt, exiting environment loop")

            if len(transitions) == 0:
                logger.warning("Trajectory is empty!")
                return

            trajectory = jax.tree.map(lambda *xs: jnp.stack(xs), *transitions)

            reward_state = get_rewards(
                trajectory=trajectory,
                rewards=constants.rewards,
                rewards_carry=env_states.reward_carry,
                rollout_length_steps=self.rollout_length_steps,
                curriculum_level=env_states.curriculum_state.level,
                clip_min=self.config.reward_clip_min,
                clip_max=self.config.reward_clip_max,
            )

            if save_path is not None:
                self._log_logged_trajectory_graphs(
                    logged_traj=LoggedTrajectory(
                        trajectory=trajectory,
                        rewards=reward_state,
                        metrics=xax.FrozenDict({}),
                    ),
                    log_callback=lambda name, value, _: value.save(save_path / f"{name}.png"),
                )

                self._save_viewer_video(frames, save_path)

    def _save_viewer_video(self, frames: list[np.ndarray], save_path: Path) -> None:
        fps = round(1 / self.config.ctrl_dt)
        vid_save_path = save_path / ("render.mp4" if self.config.run_viewer_save_video else "render.gif")

        match vid_save_path.suffix.lower():
            case ".mp4":
                try:
                    import imageio.v2 as imageio

                except ImportError as err:
                    raise RuntimeError(
                        "Failed to save video - note that saving .mp4 videos with imageio usually "
                        "requires the FFMPEG backend, which can be installed using `pip install "
                        "'imageio[ffmpeg]'`. Note that this also requires FFMPEG to be installed in "
                        "your system."
                    ) from err

                try:
                    with imageio.get_writer(vid_save_path, mode="I", fps=fps) as writer:
                        for frame in frames:
                            writer.append_data(frame)

                except Exception as e:
                    raise RuntimeError(
                        "Failed to save video - note that saving .mp4 videos with imageio usually "
                        "requires the FFMPEG backend, which can be installed using `pip install "
                        "'imageio[ffmpeg]'`. Note that this also requires FFMPEG to be installed in "
                        "your system."
                    ) from e

            case ".gif":
                images = [Image.fromarray(frame) for frame in frames]
                images[0].save(
                    vid_save_path,
                    save_all=True,
                    append_images=images[1:],
                    duration=int(1000 / fps),
                    loop=0,
                )

            case _:
                raise ValueError(f"Unsupported file extension: {vid_save_path.suffix}. Expected .mp4 or .gif")

        logger.log(xax.LOG_STATUS, "Rendered trajectory visuals to %s", save_path)

    def _get_constants(
        self,
        *,
        mj_model: mujoco.MjModel,
        physics_model: PhysicsModel,
        model_statics: tuple[PyTree, ...],
        argmax_action: bool,
    ) -> RolloutConstants:
        if len(model_statics) < 1:
            raise ValueError("No models found")

        metadata = self.get_mujoco_model_metadata(mj_model)
        engine = self.get_engine(physics_model, metadata)
        observations = self.get_observations(physics_model)
        commands = self.get_commands(physics_model)
        rewards_terms = self.get_rewards(physics_model)
        terminations = self.get_terminations(physics_model)
        curriculum = self.get_curriculum(physics_model)

        return RolloutConstants(
            model_statics=model_statics,
            engine=engine,
            observations=tuple(observations),
            commands=tuple(commands),
            rewards=tuple(rewards_terms),
            terminations=tuple(terminations),
            curriculum=curriculum,
            argmax_action=argmax_action,
            aux_constants=xax.FrozenDict({}),
        )

    def _get_shared_state(
        self,
        *,
        mj_model: mujoco.MjModel,
        physics_model: PhysicsModel,
        model_arrs: tuple[PyTree, ...],
    ) -> RolloutSharedState:
        if len(model_arrs) < 1:
            raise ValueError("No models found")

        return RolloutSharedState(
            physics_model=physics_model,
            model_arrs=model_arrs,
            aux_values=xax.FrozenDict({}),
        )

    def _get_env_state(
        self,
        *,
        rng: PRNGKeyArray,
        rollout_constants: RolloutConstants,
        mj_model: mujoco.MjModel,
        physics_model: PhysicsModel,
        randomizers: Collection[PhysicsRandomizer],
    ) -> RolloutEnvState:
        rng, carry_rng, command_rng, rand_rng, rollout_rng, curriculum_rng, reward_rng = jax.random.split(rng, 7)

        # Vectorize across N environments for MJX models, use single model for Mujoco.
        # TODO (for a later refactor): just one code path and vmap on the outside.
        if isinstance(physics_model, mjx.Model):
            # Defines the vectorized initialization functions.
            carry_fn = jax.vmap(self.get_initial_model_carry, in_axes=0)
            command_fn = jax.vmap(get_initial_commands, in_axes=(0, 0, None, 0))
            reward_carry_fn = jax.vmap(get_initial_reward_carry, in_axes=(0, None))
            obs_carry_fn = jax.vmap(get_initial_obs_carry, in_axes=(0, None))

            # Gets the initial curriculum state.
            curriculum_state = jax.vmap(rollout_constants.curriculum.get_initial_state, in_axes=0)(
                jax.random.split(curriculum_rng, self.config.num_envs)
            )

            # Gets the per-environment randomizations.
            randomization_fn = jax.vmap(apply_randomizations, in_axes=(None, None, None, 0, 0))
            randomization_dict, physics_state = randomization_fn(
                physics_model,
                rollout_constants.engine,
                randomizers,
                curriculum_state.level,
                jax.random.split(rand_rng, self.config.num_envs),
            )

            return RolloutEnvState(
                commands=command_fn(
                    jax.random.split(command_rng, self.config.num_envs),
                    physics_state.data,
                    rollout_constants.commands,
                    curriculum_state.level,
                ),
                physics_state=physics_state,
                randomization_dict=randomization_dict,
                model_carry=carry_fn(jax.random.split(carry_rng, self.config.num_envs)),
                reward_carry=reward_carry_fn(
                    jax.random.split(reward_rng, self.config.num_envs), rollout_constants.rewards
                ),
                obs_carry=obs_carry_fn(
                    jax.random.split(carry_rng, self.config.num_envs), rollout_constants.observations
                ),
                curriculum_state=curriculum_state,
                rng=jax.random.split(rollout_rng, self.config.num_envs),
            )

        else:
            # Gets the initial curriculum state.
            curriculum_state = rollout_constants.curriculum.get_initial_state(curriculum_rng)

            # Gets the environment randomizations.
            randomization_dict, physics_state = apply_randomizations(
                physics_model,
                rollout_constants.engine,
                randomizers,
                curriculum_state.level,
                rand_rng,
            )

            return RolloutEnvState(
                commands=get_initial_commands(
                    command_rng,
                    physics_state.data,
                    rollout_constants.commands,
                    curriculum_state.level,
                ),
                physics_state=physics_state,
                randomization_dict=randomization_dict,
                model_carry=self.get_initial_model_carry(carry_rng),
                reward_carry=get_initial_reward_carry(reward_rng, rollout_constants.rewards),
                obs_carry=get_initial_obs_carry(rng=carry_rng, observations=rollout_constants.observations),
                curriculum_state=curriculum_state,
                rng=rollout_rng,
            )

    def collect_dataset(
        self,
        num_batches: int,
        save_path: str | Path | None = None,
        argmax_action: bool = False,
    ) -> None:
        """Collects a dataset of state-action pairs by running the environment loop.

        Args:
            num_batches: The number of batches to collect at a time.
            save_path: Where to save the dataset; if not specified, will save
                to the experimental directory.
            argmax_action: If set, get the argmax action, otherwise sample
                randomly from the model.
        """
        with self:
            rng = self.prng_key()
            self.set_loggers()

            if xax.is_master():
                Thread(target=self.log_state, daemon=True).start()

            # Loads the Mujoco model and logs some information about it.
            mj_model: PhysicsModel = self.get_mujoco_model()
            mj_model = self.set_mujoco_model_opts(mj_model)
            mujoco_info = OmegaConf.to_yaml(DictConfig(self.get_mujoco_model_info(mj_model)))
            self.logger.log_file("mujoco_info.yaml", mujoco_info)

            mjx_model = self.get_mjx_model(mj_model)
            randomizations = self.get_physics_randomizers(mjx_model)

            rng, model_rng = jax.random.split(rng)
            models, state = self.load_initial_state(model_rng, load_optimizer=False)

            # Partitions the models into mutable and static parts.
            model_arrs, model_statics = (
                tuple(models)
                for models in zip(
                    *(eqx.partition(model, self.model_partition_fn) for model in models),
                    strict=True,
                )
            )

            if save_path is None:
                save_path = self.exp_dir / f"dataset_{state.num_steps}.npz"

            rollout_constants = self._get_constants(
                mj_model=mj_model,
                physics_model=mjx_model,
                model_statics=model_statics,
                argmax_action=argmax_action,
            )
            rollout_env_state = self._get_env_state(
                rng=rng,
                rollout_constants=rollout_constants,
                mj_model=mj_model,
                physics_model=mjx_model,
                randomizers=randomizations,
            )
            rollout_shared_state = self._get_shared_state(
                mj_model=mj_model,
                physics_model=mjx_model,
                model_arrs=model_arrs,
            )

            state = self.on_training_start(state)

            @xax.jit()
            def get_batch(
                rollout_env_state: RolloutEnvState,
            ) -> tuple[Trajectory, RewardState, RolloutEnvState]:
                vmapped_unroll = jax.vmap(self._single_unroll, in_axes=(None, 0, None))
                return vmapped_unroll(rollout_constants, rollout_env_state, rollout_shared_state)

            with TrajectoryDataset.writer(save_path, num_batches * self.batch_size) as writer:
                for _ in tqdm.trange(num_batches):
                    trajectories, rewards, rollout_env_state = get_batch(rollout_env_state)

                    # Splits trajectories and rewards into a list of `batch_size` samples.
                    for i in range(0, len(trajectories.done), self.batch_size):
                        trajectory = jax.tree.map(lambda x, index=i: x[index], trajectories)
                        reward = jax.tree.map(lambda x, index=i: x[index], rewards)
                        writer.write(trajectory, reward)

            logger.info("Saved dataset to %s", save_path)

    def initialize_rl_training(
        self,
        mj_model: PhysicsModel,
        rng: PRNGKeyArray,
    ) -> tuple[RLLoopConstants, RLLoopCarry, xax.State]:
        # Gets the model and optimizer variables.
        rng, model_rng = jax.random.split(rng)
        models, optimizers, opt_states, state = self.load_initial_state(model_rng, load_optimizer=True)

        # Logs model and optimizer information.
        for i, (model, opt_state) in enumerate(zip(models, opt_states, strict=True), 1):
            suffix = f" {i}" if len(models) > 1 else ""
            model_size = xax.get_pytree_param_count(model)
            opt_state_size = xax.get_pytree_param_count(opt_state)
            logger.log(xax.LOG_PING, "Model%s size: %s parameters", suffix, f"{model_size:,}")
            logger.log(xax.LOG_PING, "Optimizer%s size: %s parameters", suffix, f"{opt_state_size:,}")

        # Partitions the models into mutable and static parts.
        model_arrs, model_statics = (
            tuple(models)
            for models in zip(
                *(eqx.partition(model, self.model_partition_fn) for model in models),
                strict=True,
            )
        )

        # Loads the MJX model, and initializes the loop variables.
        mjx_model = self.get_mjx_model(mj_model)
        randomizers = self.get_physics_randomizers(mjx_model)

        constants = RLLoopConstants(
            optimizer=tuple(optimizers),
            constants=self._get_constants(
                mj_model=mj_model,
                physics_model=mjx_model,
                model_statics=model_statics,
                argmax_action=False,
            ),
        )

        carry = RLLoopCarry(
            opt_state=tuple(opt_states),
            env_states=self._get_env_state(
                rng=rng,
                rollout_constants=constants.constants,
                mj_model=mj_model,
                physics_model=mjx_model,
                randomizers=randomizers,
            ),
            shared_state=self._get_shared_state(
                mj_model=mj_model,
                physics_model=mjx_model,
                model_arrs=model_arrs,
            ),
        )

        return constants, carry, state

    def _save(
        self,
        constants: RLLoopConstants,
        carry: RLLoopCarry,
        state: xax.State,
    ) -> None:
        model = eqx.combine(carry.shared_state.model_arrs, constants.constants.model_statics)
        self.save_checkpoint(models=[model], optimizers=constants.optimizer, opt_states=carry.opt_state, state=state)

    def run_training(self) -> None:
        """Wraps the training loop and provides clean XAX integration."""
        with self:
            rng = self.prng_key()
            self.set_loggers()

            if xax.is_master():
                Thread(target=self.log_state, daemon=True).start()

            # Loads the Mujoco model and logs some information about it.
            mj_model: PhysicsModel = self.get_mujoco_model()
            mj_model = self.set_mujoco_model_opts(mj_model)
            mujoco_info = OmegaConf.to_yaml(DictConfig(self.get_mujoco_model_info(mj_model)))
            self.logger.log_file("mujoco_info.yaml", mujoco_info)

            constants, carry, state = self.initialize_rl_training(mj_model, rng)

            # Creates the markers.
            markers = self.get_markers(
                commands=constants.constants.commands,
                observations=constants.constants.observations,
                rewards=constants.constants.rewards,
            )

            # Creates the viewer.
            small_viewer = get_viewer(
                mj_model=mj_model,
                config=self.config,
                mode="offscreen",
                is_small=True,
            )
            full_viewer = get_viewer(
                mj_model=mj_model,
                config=self.config,
                mode="offscreen",
                is_small=False,
            )

            state = self.on_training_start(state)

            def on_exit() -> None:
                self._save(constants=constants, carry=carry, state=state)

            # Handle user-defined interrupts during the training loop.
            self.add_signal_handler(on_exit, signal.SIGUSR1, signal.SIGTERM)

            is_first_step = True

            try:
                while not self.is_training_over(state):
                    # Runs the training loop.
                    with xax.ContextTimer() as timer:
                        valid_step = self.valid_step_timer(state)

                        state = state.replace(
                            phase="valid" if valid_step else "train",
                        )

                        state = self.on_step_start(state)

                        rng, update_rng = jax.random.split(rng)
                        carry, metrics, logged_traj = self._rl_train_loop_step(
                            carry=carry,
                            constants=constants,
                            state=state,
                            rng=update_rng,
                        )

                        if self.config.profile_memory:
                            carry = jax.block_until_ready(carry)
                            metrics = jax.block_until_ready(metrics)
                            logged_traj = jax.block_until_ready(logged_traj)
                            jax.profiler.save_device_memory_profile(self.exp_dir / "train_loop_step.prof")

                        self.log_train_metrics(metrics)
                        self.log_state_timers(state)

                        if self.should_checkpoint(state):
                            self._save(constants=constants, carry=carry, state=state)

                        state = self.on_step_end(state)

                        # Updates the step and sample counts.
                        num_steps = self.config.epochs_per_log_step
                        num_samples = self.rollout_num_samples * self.config.epochs_per_log_step

                        if valid_step:
                            state = state.replace(
                                num_valid_steps=state.num_valid_steps + num_steps,
                                num_valid_samples=state.num_valid_samples + num_samples,
                            )

                            full_size_render = state.num_valid_steps.item() % self.config.render_full_every_n_steps == 0
                            if full_size_render:
                                self._log_logged_trajectory_video(
                                    logged_traj=logged_traj,
                                    markers=markers,
                                    viewer=full_viewer,
                                    key="trajectory",
                                )
                                self._log_logged_trajectory_graphs(
                                    logged_traj=logged_traj,
                                    log_callback=lambda key, value, namespace: self.logger.log_image(
                                        key=key, value=value, namespace=namespace
                                    ),
                                )
                            else:
                                self._log_logged_trajectory_video(
                                    logged_traj=logged_traj,
                                    markers=markers,
                                    viewer=small_viewer,
                                    key="trajectory_small",
                                )

                        else:
                            state = state.replace(
                                num_steps=state.num_steps + num_steps,
                                num_samples=state.num_samples + num_samples,
                            )

                        self.write_logs(state)

                    # Update  state with the elapsed time.
                    elapsed_time = timer.elapsed_time
                    if valid_step:
                        state = state.replace(
                            valid_elapsed_time_s=state.valid_elapsed_time_s + elapsed_time,
                        )
                    else:
                        state = state.replace(
                            elapsed_time_s=state.elapsed_time_s + elapsed_time,
                        )

                    if is_first_step:
                        is_first_step = False
                        logger.log(
                            xax.LOG_STATUS,
                            "First step time: %s",
                            xax.format_timedelta(datetime.timedelta(seconds=elapsed_time), short=True),
                        )

                # Save the checkpoint when done.
                self._save(constants=constants, carry=carry, state=state)

            except xax.TrainingFinishedError:
                if xax.is_master():
                    msg = f"Finished training after {state.num_steps}steps and {state.num_samples} samples"
                    xax.show_info(msg, important=True)
                self._save(constants=constants, carry=carry, state=state)

            except (KeyboardInterrupt, bdb.BdbQuit):
                if xax.is_master():
                    xax.show_info("Interrupted training", important=True)

            except BaseException:
                exception_tb = textwrap.indent(xax.highlight_exception_message(traceback.format_exc()), "  ")
                sys.stdout.write(f"Caught exception during training loop:\n\n{exception_tb}\n")
                sys.stdout.flush()
                self._save(constants=constants, carry=carry, state=state)

            finally:
                state = self.on_training_end(state)



# mypy: disable-error-code="override"
"""Defines simple task for training a walking policy for the default humanoid using an RNN actor."""

from dataclasses import dataclass
from typing import Generic, TypeVar

import distrax
import equinox as eqx
import jax
import jax.numpy as jnp
import xax
from jaxtyping import Array, PRNGKeyArray

import ksim

from .walking import (
    NUM_INPUTS,
    NUM_JOINTS,
    HumanoidWalkingTask,
    HumanoidWalkingTaskConfig,
)


class DefaultHumanoidRNNActor(eqx.Module):
    """RNN-based actor for the walking task."""

    input_proj: eqx.nn.Linear
    rnns: tuple[eqx.nn.GRUCell, ...]
    output_proj: eqx.nn.Linear
    num_inputs: int = eqx.static_field()
    num_outputs: int = eqx.static_field()
    min_std: float = eqx.static_field()
    max_std: float = eqx.static_field()
    var_scale: float = eqx.static_field()

    def __init__(
        self,
        key: PRNGKeyArray,
        *,
        num_inputs: int,
        num_outputs: int,
        min_std: float,
        max_std: float,
        var_scale: float,
        hidden_size: int,
        depth: int,
    ) -> None:
        # Project input to hidden size
        key, input_proj_key = jax.random.split(key)
        self.input_proj = eqx.nn.Linear(
            in_features=num_inputs,
            out_features=hidden_size,
            key=input_proj_key,
        )

        # Create RNN layer
        key, rnn_key = jax.random.split(key)
        self.rnns = tuple(
            [
                eqx.nn.GRUCell(
                    input_size=hidden_size,
                    hidden_size=hidden_size,
                    key=rnn_key,
                )
                for _ in range(depth)
            ]
        )

        # Project to output
        self.output_proj = eqx.nn.Linear(
            in_features=hidden_size,
            out_features=num_outputs * 2,
            key=key,
        )

        self.num_inputs = num_inputs
        self.num_outputs = num_outputs
        self.min_std = min_std
        self.max_std = max_std
        self.var_scale = var_scale

    def forward(self, obs_n: Array, carry: Array) -> tuple[distrax.Distribution, Array]:
        x_n = self.input_proj(obs_n)
        out_carries = []
        for i, rnn in enumerate(self.rnns):
            x_n = rnn(x_n, carry[i])
            out_carries.append(x_n)
        out_n = self.output_proj(x_n)

        # Converts the output to a distribution.
        mean_n = out_n[..., : self.num_outputs]
        std_n = out_n[..., self.num_outputs :]

        # Softplus and clip to ensure positive standard deviations.
        std_n = jnp.clip((jax.nn.softplus(std_n) + self.min_std) * self.var_scale, max=self.max_std)

        dist_n = distrax.Normal(mean_n, std_n)
        return dist_n, jnp.stack(out_carries, axis=0)


class DefaultHumanoidRNNCritic(eqx.Module):
    """RNN-based critic for the walking task."""

    input_proj: eqx.nn.Linear
    rnns: tuple[eqx.nn.GRUCell, ...]
    output_proj: eqx.nn.Linear

    def __init__(
        self,
        key: PRNGKeyArray,
        *,
        num_inputs: int,
        hidden_size: int,
        depth: int,
    ) -> None:
        num_outputs = 1

        # Project input to hidden size
        key, input_proj_key = jax.random.split(key)
        self.input_proj = eqx.nn.Linear(
            in_features=num_inputs,
            out_features=hidden_size,
            key=input_proj_key,
        )

        # Create RNN layer
        key, rnn_key = jax.random.split(key)
        self.rnns = tuple(
            [
                eqx.nn.GRUCell(
                    input_size=hidden_size,
                    hidden_size=hidden_size,
                    key=rnn_key,
                )
                for _ in range(depth)
            ]
        )

        # Project to output
        self.output_proj = eqx.nn.Linear(
            in_features=hidden_size,
            out_features=num_outputs,
            key=key,
        )

    def forward(self, obs_n: Array, carry: Array) -> tuple[Array, Array]:
        x_n = self.input_proj(obs_n)
        out_carries = []
        for i, rnn in enumerate(self.rnns):
            x_n = rnn(x_n, carry[i])
            out_carries.append(x_n)
        out_n = self.output_proj(x_n)

        return out_n, jnp.stack(out_carries, axis=0)


class DefaultHumanoidRNNModel(eqx.Module):
    actor: DefaultHumanoidRNNActor
    critic: DefaultHumanoidRNNCritic

    def __init__(
        self,
        key: PRNGKeyArray,
        *,
        min_std: float,
        max_std: float,
        num_inputs: int,
        num_joints: int,
        hidden_size: int,
        depth: int,
    ) -> None:
        self.actor = DefaultHumanoidRNNActor(
            key,
            num_inputs=num_inputs,
            num_outputs=num_joints,
            min_std=min_std,
            max_std=max_std,
            var_scale=0.5,
            hidden_size=hidden_size,
            depth=depth,
        )
        self.critic = DefaultHumanoidRNNCritic(
            key,
            num_inputs=num_inputs,
            hidden_size=hidden_size,
            depth=depth,
        )


@dataclass
class HumanoidWalkingRNNTaskConfig(HumanoidWalkingTaskConfig):
    pass


Config = TypeVar("Config", bound=HumanoidWalkingRNNTaskConfig)


class HumanoidWalkingRNNTask(HumanoidWalkingTask[Config], Generic[Config]):
    def get_model(self, key: PRNGKeyArray) -> DefaultHumanoidRNNModel:
        return DefaultHumanoidRNNModel(
            key,
            num_inputs=NUM_INPUTS,
            num_joints=NUM_JOINTS,
            min_std=0.01,
            max_std=1.0,
            hidden_size=self.config.hidden_size,
            depth=self.config.depth,
        )

    def run_actor(
        self,
        model: DefaultHumanoidRNNActor,
        observations: xax.FrozenDict[str, Array],
        commands: xax.FrozenDict[str, Array],
        carry: Array,
    ) -> tuple[distrax.Distribution, Array]:
        timestep_1 = observations["timestep_observation"]
        dh_joint_pos_j = observations["joint_position_observation"]
        dh_joint_vel_j = observations["joint_velocity_observation"]
        com_inertia_n = observations["center_of_mass_inertia_observation"]
        com_vel_n = observations["center_of_mass_velocity_observation"]
        # imu_acc_3 = observations["sensor_observation_imu_acc"]
        # imu_gyro_3 = observations["sensor_observation_imu_gyro"]
        proj_grav_3 = observations["projected_gravity_observation"]
        act_frc_obs_n = observations["actuator_force_observation"]
        base_pos_3 = observations["base_position_observation"]
        base_quat_4 = observations["base_orientation_observation"]
        lin_vel_obs_3 = observations["base_linear_velocity_observation"]
        ang_vel_obs_3 = observations["base_angular_velocity_observation"]
        joystick_cmd_1 = commands["joystick_command"]
        joystick_cmd_ohe_6 = jax.nn.one_hot(joystick_cmd_1, num_classes=6).squeeze(-2)

        obs_n = jnp.concatenate(
            [
                jnp.cos(timestep_1),  # 1
                jnp.sin(timestep_1),  # 1
                dh_joint_pos_j,  # NUM_JOINTS
                dh_joint_vel_j / 10.0,  # NUM_JOINTS
                com_inertia_n,  # 160
                com_vel_n,  # 96
                proj_grav_3,  # 3
                act_frc_obs_n / 100.0,  # NUM_JOINTS
                base_pos_3,  # 3
                base_quat_4,  # 4
                lin_vel_obs_3,  # 3
                ang_vel_obs_3,  # 3
                joystick_cmd_ohe_6,  # 6
            ],
            axis=-1,
        )

        return model.forward(obs_n, carry)

    def run_critic(
        self,
        model: DefaultHumanoidRNNCritic,
        observations: xax.FrozenDict[str, Array],
        commands: xax.FrozenDict[str, Array],
        carry: Array,
    ) -> tuple[Array, Array]:
        timestep_1 = observations["timestep_observation"]
        dh_joint_pos_j = observations["joint_position_observation"]
        dh_joint_vel_j = observations["joint_velocity_observation"]
        com_inertia_n = observations["center_of_mass_inertia_observation"]
        com_vel_n = observations["center_of_mass_velocity_observation"]
        # imu_acc_3 = observations["sensor_observation_imu_acc"]
        # imu_gyro_3 = observations["sensor_observation_imu_gyro"]
        proj_grav_3 = observations["projected_gravity_observation"]
        act_frc_obs_n = observations["actuator_force_observation"]
        base_pos_3 = observations["base_position_observation"]
        base_quat_4 = observations["base_orientation_observation"]
        lin_vel_obs_3 = observations["base_linear_velocity_observation"]
        ang_vel_obs_3 = observations["base_angular_velocity_observation"]
        joystick_cmd_1 = commands["joystick_command"]
        joystick_cmd_ohe_6 = jax.nn.one_hot(joystick_cmd_1, num_classes=6).squeeze(-2)

        obs_n = jnp.concatenate(
            [
                jnp.cos(timestep_1),  # 1
                jnp.sin(timestep_1),  # 1
                dh_joint_pos_j,  # NUM_JOINTS
                dh_joint_vel_j / 10.0,  # NUM_JOINTS
                com_inertia_n,  # 160
                com_vel_n,  # 96
                proj_grav_3,  # 3
                act_frc_obs_n / 100.0,  # NUM_JOINTS
                base_pos_3,  # 3
                base_quat_4,  # 4
                lin_vel_obs_3,  # 3
                ang_vel_obs_3,  # 3
                joystick_cmd_ohe_6,  # 6
            ],
            axis=-1,
        )

        return model.forward(obs_n, carry)

    def get_ppo_variables(
        self,
        model: DefaultHumanoidRNNModel,
        trajectory: ksim.Trajectory,
        model_carry: tuple[Array, Array],
        rng: PRNGKeyArray,
    ) -> tuple[ksim.PPOVariables, tuple[Array, Array]]:
        def scan_fn(
            actor_critic_carry: tuple[Array, Array],
            transition: ksim.Trajectory,
        ) -> tuple[tuple[Array, Array], ksim.PPOVariables]:
            actor_carry, critic_carry = actor_critic_carry
            actor_dist, next_actor_carry = self.run_actor(
                model=model.actor,
                observations=transition.obs,
                commands=transition.command,
                carry=actor_carry,
            )
            log_probs = actor_dist.log_prob(transition.action)
            assert isinstance(log_probs, Array)
            value, next_critic_carry = self.run_critic(
                model=model.critic,
                observations=transition.obs,
                commands=transition.command,
                carry=critic_carry,
            )

            transition_ppo_variables = ksim.PPOVariables(
                log_probs=log_probs,
                values=value.squeeze(-1),
            )

            next_carry = jax.tree.map(
                lambda x, y: jnp.where(transition.done, x, y),
                self.get_initial_model_carry(rng),
                (next_actor_carry, next_critic_carry),
            )

            return next_carry, transition_ppo_variables

        next_model_carry, ppo_variables = jax.lax.scan(scan_fn, model_carry, trajectory)

        return ppo_variables, next_model_carry

    def get_initial_model_carry(self, rng: PRNGKeyArray) -> tuple[Array, Array]:
        return (
            jnp.zeros(shape=(self.config.depth, self.config.hidden_size)),
            jnp.zeros(shape=(self.config.depth, self.config.hidden_size)),
        )

    def sample_action(
        self,
        model: DefaultHumanoidRNNModel,
        model_carry: tuple[Array, Array],
        physics_model: ksim.PhysicsModel,
        physics_state: ksim.PhysicsState,
        observations: xax.FrozenDict[str, Array],
        commands: xax.FrozenDict[str, Array],
        rng: PRNGKeyArray,
        argmax: bool,
    ) -> ksim.Action:
        actor_carry_in, critic_carry_in = model_carry

        # Runs the actor model to get the action distribution.
        action_dist_j, actor_carry = self.run_actor(
            model=model.actor,
            observations=observations,
            commands=commands,
            carry=actor_carry_in,
        )

        action_j = action_dist_j.mode() if argmax else action_dist_j.sample(seed=rng)

        return ksim.Action(
            action=action_j,
            carry=(actor_carry, critic_carry_in),
            aux_outputs=None,
        )


if __name__ == "__main__":
    # To run training, use the following command:
    #   python -m examples.walking_rnn
    # To visualize the environment, use the following command:
    #   python -m examples.walking_rnn run_model_viewer=True
    HumanoidWalkingRNNTask.launch(
        HumanoidWalkingRNNTaskConfig(
            # Training parameters.
            num_envs=2048,
            batch_size=256,
            num_passes=2,
            epochs_per_log_step=1,
            rollout_length_seconds=8.0,
            # Simulation parameters.
            dt=0.002,
            ctrl_dt=0.02,
            iterations=3,
            ls_iterations=5,
            max_action_latency=0.01,
        ),
    )



"""Reference motion utilities."""

__all__ = [
    "MotionReferenceMapping",
    "MotionReferenceData",
    "get_local_xpos",
    "get_local_reference_pos",
    "local_to_absolute",
    "get_reference_joint_id",
    "get_body_id",
    "get_reference_joint_ids",
    "get_body_ids",
    "visualize_reference_points",
    "visualize_reference_motion",
    "get_reference_cartesian_poses",
    "generate_reference_motion",
]

import argparse
import logging
import sys
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, Self, Sequence, cast

import bvhio
import colorlogging
import glm
import jax
import jax.numpy as jnp
import mujoco
import numpy as np
import xax
from bvhio.lib.hierarchy import Joint as BvhioJoint
from jaxtyping import Array
from omegaconf import MISSING, DictConfig, ListConfig, OmegaConf
from omegaconf.errors import ConfigKeyError, MissingMandatoryValue
from scipy.optimize import least_squares
from scipy.spatial.transform import Rotation as R

from ksim.viewer import GlfwMujocoViewer

logger = logging.getLogger(__name__)


@dataclass
class MotionReferenceMapping:
    reference_joint_name: str
    mj_body_name: str


@dataclass(frozen=True)
class MotionReferenceData:
    """Stores reference motion data (qpos and Cartesian poses)."""

    qpos: xax.HashableArray  # Shape: [T, nq]
    qvel: xax.HashableArray  # Shape: [T, nq]
    cartesian_poses: xax.FrozenDict[int, xax.HashableArray]  # Dict: body_id -> [T, 3]
    ctrl_dt: float

    @property
    def num_frames(self) -> int:
        """Returns the total number of frames in the reference motion."""
        return self.qpos.array.shape[0]

    def get_qpos_at_step(self, step: int | Array) -> Array:
        """Gets the reference qpos at a specific step (or steps)."""
        frame_index = step % self.num_frames
        return jnp.take(self.qpos.array, frame_index, axis=0)

    def get_qvel_at_step(self, step: int | Array) -> Array:
        """Gets the reference qvel at a specific step (or steps)."""
        frame_index = step % self.num_frames
        return jnp.take(self.qvel.array, frame_index, axis=0)

    def get_cartesian_pose_at_step(self, step: int | Array) -> xax.FrozenDict[int, Array]:
        """Gets the reference Cartesian pose at a specific step (or steps)."""
        frame_index = step % self.num_frames
        return jax.tree.map(
            lambda hashable_arr: jnp.take(hashable_arr.array, frame_index, axis=0), self.cartesian_poses
        )

    def get_qpos_at_time(self, time: float | Array) -> Array:
        """Gets the reference qpos closest to a specific time."""
        step = jnp.int32(jnp.round(time / self.ctrl_dt))
        return self.get_qpos_at_step(step)

    def get_cartesian_pose_at_time(self, time: float | Array) -> xax.FrozenDict[int, Array]:
        """Gets the reference Cartesian pose closest to a specific time."""
        step = jnp.int32(jnp.round(time / self.ctrl_dt))
        return self.get_cartesian_pose_at_step(step)

    def save(self, path: str | Path) -> None:
        """Saves the MotionReferenceData to a .npz file."""
        save_path = Path(path)
        save_path.parent.mkdir(parents=True, exist_ok=True)

        body_ids = np.array(list(self.cartesian_poses.keys()))
        pose_arrays = [pose.array for pose in self.cartesian_poses.values()]

        pose_arrays_np = [np.asarray(arr) for arr in pose_arrays]
        cartesian_poses_stacked = np.stack(pose_arrays_np, axis=0)  # Shape [num_bodies, T, 3]

        np.savez(
            save_path,
            qpos=np.asarray(self.qpos.array),
            qvel=np.asarray(self.qvel.array),
            cartesian_pose_body_ids=body_ids,
            cartesian_poses_stacked=cartesian_poses_stacked,
            ctrl_dt=np.array(self.ctrl_dt),
        )
        logger.info("Saved reference motion data to %s", save_path)

    @classmethod
    def load(cls, path: str | Path) -> Self:
        """Loads MotionReferenceData from a .npz file."""
        load_path = Path(path)
        if not load_path.exists():
            raise FileNotFoundError(f"Reference motion file not found: {load_path}")

        logger.info("Loading reference motion data from %s", load_path)
        data = np.load(load_path)

        qpos = xax.HashableArray(jnp.array(data["qpos"]))
        qvel = xax.HashableArray(jnp.array(data["qvel"]))
        ctrl_dt = float(data["ctrl_dt"].item())

        # Reconstruct cartesian poses dictionary
        body_ids = data["cartesian_pose_body_ids"]
        cartesian_poses_stacked = data["cartesian_poses_stacked"]
        cartesian_poses: xax.FrozenDict[int, xax.HashableArray] = xax.FrozenDict(
            {
                int(body_id): xax.HashableArray(jnp.array(cartesian_poses_stacked[i]))
                for i, body_id in enumerate(body_ids)
            }
        )

        return cls(
            qpos=qpos,
            qvel=qvel,
            cartesian_poses=cartesian_poses,
            ctrl_dt=ctrl_dt,
        )


def _add_reference_marker_to_scene(
    scene: mujoco.MjvScene,
    *,
    pos: np.ndarray,
    color: np.ndarray,
    scale: np.ndarray,
    label: str | None = None,
) -> None:
    """Adds a sphere to the scene.

    Args:
        scene: The MjvScene to add the sphere to
        pos: The position of the sphere
        color: The color of the sphere
        scale: The scale of the sphere
        label: The label of the sphere
    """
    if scene.ngeom >= scene.maxgeom:
        raise ValueError("Scene is full")

    g = scene.geoms[scene.ngeom]
    g.type = mujoco.mjtGeom.mjGEOM_SPHERE
    g.size[:] = scale
    g.pos[:] = pos
    g.mat[:] = np.eye(3)
    g.rgba[:] = color
    g.label = label or ""
    g.dataid = -1
    g.objtype = mujoco.mjtObj.mjOBJ_UNKNOWN
    g.objid = -1
    g.category = mujoco.mjtCatBit.mjCAT_DECOR
    g.emission = 0
    g.specular = 0.5
    g.shininess = 0.5

    scene.ngeom += 1


def get_local_xpos(xpos: np.ndarray | jax.Array, body_id: int, base_id: int) -> np.ndarray | jax.Array:
    """Gets the cartesian pos of a point w.r.t. the base (e.g. pelvis)."""
    return xpos[body_id] - xpos[base_id]


def get_local_reference_pos(
    root: BvhioJoint,
    reference_id: int,
    reference_base_id: int,
    scaling_factor: float = 1.0,
) -> np.ndarray | jax.Array:
    """Gets the cartesian pos of a reference joint w.r.t. the base (e.g. pelvis)."""
    layout = root.layout()
    joint = layout[reference_id][0]
    reference_joint = layout[reference_base_id][0]
    ref_pos = np.array(joint.PositionWorld) - np.array(reference_joint.PositionWorld)
    return ref_pos * scaling_factor


def local_to_absolute(
    xpos: np.ndarray | jax.Array, local_pos: np.ndarray | jax.Array, base_id: int
) -> np.ndarray | jax.Array:
    """Gets the absolute xpos from a local position (for visualization)."""
    return local_pos + xpos[base_id]


def get_reference_joint_id(root: BvhioJoint, reference_joint_name: str) -> int:
    for joint, index, _ in root.layout():
        if joint.Name == reference_joint_name:
            return index
    raise ValueError(f"Joint {reference_joint_name} not found")


def get_body_id(model: mujoco.MjModel, mj_body_name: str) -> int:
    return mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_BODY, mj_body_name)


def get_reference_joint_ids(root: BvhioJoint, mappings: tuple[MotionReferenceMapping, ...]) -> tuple[int, ...]:
    return tuple([get_reference_joint_id(root, mapping.reference_joint_name) for mapping in mappings])


def get_body_ids(model: mujoco.MjModel, mappings: tuple[MotionReferenceMapping, ...]) -> tuple[int, ...]:
    return tuple([get_body_id(model, mapping.mj_body_name) for mapping in mappings])


def visualize_reference_points(
    model: mujoco.MjModel,
    base_id: int,
    reference_motion: xax.FrozenDict[int, np.ndarray],
) -> None:
    """Animates the model and adds real geoms to the scene for each joint position.

    Args:
        model: The Mujoco model
        base_id: The ID of the Mujoco base
        reference_motion: The reference motion (if root and reference_base_id are None)
        display_names: Whether to display the names of the bodies / reference joints
    """
    total_frames = list(reference_motion.values())[0].shape[0]
    data = mujoco.MjData(model)

    with mujoco.viewer.launch_passive(model, data) as viewer:
        frame = 0
        while viewer.is_running():
            frame = frame % total_frames

            scene = viewer.user_scn
            scene.ngeom = 0  # Clear previous geoms

            # Showing default humanoid geoms for reference
            for body_id, reference_poses in reference_motion.items():
                agent_local_pos = get_local_xpos(data.xpos, body_id, base_id)
                agent_xpos = local_to_absolute(data.xpos, agent_local_pos, base_id)
                assert isinstance(agent_xpos, np.ndarray)
                _add_reference_marker_to_scene(
                    scene,
                    pos=agent_xpos,
                    color=np.array([1, 0, 0, 1]),
                    scale=np.array([0.08, 0.08, 0.08]),
                )

                reference_local_pos = reference_poses[frame]
                reference_xpos = local_to_absolute(data.xpos, reference_local_pos, base_id)
                assert isinstance(reference_xpos, np.ndarray)
                _add_reference_marker_to_scene(
                    scene,
                    pos=reference_xpos,
                    color=np.array([0, 1, 0, 1]),
                    scale=np.array([0.08, 0.08, 0.08]),
                )

            viewer.sync()
            time.sleep(0.01)
            frame += 1


def visualize_reference_motion(
    model: mujoco.MjModel,
    reference_qpos: np.ndarray,
    cartesian_motion: xax.FrozenDict[int, np.ndarray],
    mj_base_id: int,
) -> None:
    data = mujoco.MjData(model)
    viewer = GlfwMujocoViewer(model, data, mode="window", width=1024, height=768)

    # Set some nice camera parameters
    viewer.cam.distance = 3.0
    viewer.cam.azimuth = 45.0
    viewer.cam.elevation = -20.0

    # Keep the window open until you close it
    while viewer.is_alive:
        # Calculate which frame we should be on
        frame = int(data.time / model.opt.timestep) % len(reference_qpos)

        # Update the pose from reference qpos
        data.qpos = reference_qpos[frame]
        mujoco.mj_forward(model, data)

        # Clear previous geoms
        scene = viewer.scn
        scene.ngeom = 0

        def marker_callback(
            model: mujoco.MjModel,
            data: mujoco.MjData,
            scn: mujoco.MjvScene,
            frame_id: int = frame,
        ) -> None:
            # Add markers for cartesian targets
            for body_id, target_poses in cartesian_motion.items():
                target_local_pos = target_poses[frame_id]
                target_xpos = local_to_absolute(data.xpos, target_local_pos, mj_base_id)
                assert isinstance(target_xpos, np.ndarray)
                _add_reference_marker_to_scene(
                    scn,
                    pos=target_xpos,
                    color=np.array([0, 1, 0, 0.8]),
                    scale=np.array([0.05, 0.05, 0.05]),
                    label=f"target_{model.body(body_id).name}",
                )

        # Advance simulation time
        data.time += model.opt.timestep

        viewer.render(callback=marker_callback)


def get_reference_cartesian_poses(
    mappings: tuple[MotionReferenceMapping, ...],
    model: mujoco.MjModel,
    root: BvhioJoint,
    reference_base_id: int,
    root_callback: Callable[[BvhioJoint], None] | None,
    scaling_factor: float = 1.0,
    offset: np.ndarray | None = None,
) -> xax.FrozenDict[int, np.ndarray]:
    """Generates the reference motion for the given model and data.

    Args:
        mappings: The mappings of BVH joints to Mujoco bodies
        model: The Mujoco model
        root: The root of the BVH tree
        reference_base_id: The ID of the reference base (of the BVH file)
        root_callback: A callback to modify the root of the BVH tree
        scaling_factor: The scaling factor for the reference motion
        offset: The offset of the reference motion

    Returns:
        A tuple of tuples, where each tuple contains a Mujoco body ID and the target positions.
        The result will be of shape [T, 3].
    """
    if offset is None:
        offset = np.array([0.0, 0.0, 0.0])

    reference_joint_ids = get_reference_joint_ids(root, mappings)
    body_ids = get_body_ids(model, mappings)

    total_frames = len(root.layout()[0][0].Keyframes)
    reference_motion: xax.FrozenDict[int, np.ndarray] = xax.FrozenDict(
        {body_id: np.zeros((total_frames, 3)) for body_id in body_ids}
    )

    for frame in range(total_frames):
        root.loadPose(frame)

        if root_callback:
            root_callback(root)

        for body_id, reference_joint_id in zip(body_ids, reference_joint_ids, strict=False):
            ref_pos = get_local_reference_pos(root, reference_joint_id, reference_base_id, scaling_factor)
            reference_motion[body_id][frame] = ref_pos + offset

    return reference_motion


def multi_body_ik_error(
    x: np.ndarray,
    *,
    model: mujoco.MjModel,
    data: mujoco.MjData,
    mj_base_id: int,
    constrained_jnt_mask: np.ndarray,
    cartesian_pose: xax.FrozenDict[int, np.ndarray],
    neutral_qpos: np.ndarray,
    prev_qpos: np.ndarray,
    neutral_similarity_weight: float = 0.1,
    temporal_consistency_weight: float = 0.1,
) -> np.ndarray:
    """Computes the error of the multi-body IK.

    Args:
        x: Candidate qpos (without constrained joints)
        model: The Mujoco model
        data: The Mujoco data
        mj_base_id: The ID of the Mujoco base (e.g. pelvis)
        constrained_jnt_mask: In order of the qpos array, 1 = fix to neutral pos
        cartesian_pose: The Cartesian position with respect to the base
        neutral_qpos: The neutral qpos
        prev_qpos: The previous qpos
        neutral_similarity_weight: The weight of the neutral similarity term
        temporal_consistency_weight: The weight of the temporal consistency term

    Returns:
        An array of individual error terms.
    """
    new_qpos = np.copy(neutral_qpos)
    new_qpos[~constrained_jnt_mask] = x
    data.qpos = new_qpos
    mujoco.mj_forward(model, data)  # Needed to get xpos

    errors = []
    for body_id, target_pos in cartesian_pose.items():
        # The target pos is w.r.t. the base, so we need to get the local pos
        current_pos = get_local_xpos(data.xpos, body_id, mj_base_id)
        error = target_pos - current_pos
        errors.append(error)

    if neutral_qpos is not None:
        neutral_similarity_error = (data.qpos - neutral_qpos) * neutral_similarity_weight
        errors.append(neutral_similarity_error)

    if prev_qpos is not None:
        temporal_consistency_error = (data.qpos - prev_qpos) * temporal_consistency_weight
        errors.append(temporal_consistency_error)

    return np.concatenate(errors)


def solve_multi_body_ik(
    model: mujoco.MjModel,
    data: mujoco.MjData,
    mj_base_id: int,
    constrained_jnt_mask: np.ndarray,
    cartesian_pose: dict[int, np.ndarray],
    neutral_qpos: np.ndarray,
    prev_qpos: np.ndarray,
    neutral_similarity_weight: float = 0.1,
    temporal_consistency_weight: float = 0.1,
    n_restarts: int = 3,
    error_acceptance_threshold: float = 1e-4,
    ftol: float = 1e-8,
    xtol: float = 1e-8,
    max_nfev: int = 2000,
    verbose: bool = False,
) -> np.ndarray:
    """Solves the IK for the multi-body system.

    Args:
        model: The Mujoco model
        data: The Mujoco data
        mj_base_id: The ID of the Mujoco base
        constrained_jnt_mask: In order of the qpos array, 1 = fix to neutral pos
        cartesian_pose: The Cartesian position with respect to the base
        neutral_qpos: The neutral qpos
        prev_qpos: The previous qpos
        neutral_similarity_weight: The weight of the neutral similarity term
        temporal_consistency_weight: The weight of the temporal consistency term
        n_restarts: The number of random restarts to try
        error_acceptance_threshold: The threshold for the error
        ftol: The tolerance for the function value
        xtol: The tolerance for the solution
        max_nfev: The maximum number of function evaluations
        verbose: Whether to print verbose output

    Returns:
        The complete qpos array after solving inverse kinematics.
    """
    best_result = None
    best_cost = float("inf")

    for _ in range(n_restarts):
        qpos = data.qpos[~constrained_jnt_mask]
        initial_guess = qpos + np.random.normal(0, 0.1, size=qpos.shape)

        result = least_squares(
            multi_body_ik_error,
            initial_guess,
            kwargs={
                "model": model,
                "data": data,
                "mj_base_id": mj_base_id,
                "constrained_jnt_mask": constrained_jnt_mask,
                "cartesian_pose": cartesian_pose,
                "neutral_qpos": neutral_qpos,
                "prev_qpos": prev_qpos,
                "neutral_similarity_weight": neutral_similarity_weight,
                "temporal_consistency_weight": temporal_consistency_weight,
            },
            verbose=2 if verbose else 0,
            ftol=ftol,
            xtol=xtol,
            max_nfev=max_nfev,
        )

        if result.cost < best_cost:
            best_cost = result.cost
            best_result = result

        if best_cost < error_acceptance_threshold:
            break

    if best_result is None:
        raise RuntimeError("All IK attempts failed")

    qpos = np.copy(neutral_qpos)
    qpos[~constrained_jnt_mask] = best_result.x
    return qpos


def generate_reference_motion(
    model: mujoco.MjModel,
    mj_base_id: int,
    bvh_root: BvhioJoint,
    bvh_to_mujoco_names: tuple[MotionReferenceMapping, ...],
    bvh_base_id: int,
    ctrl_dt: float,
    bvh_offset: np.ndarray | None = None,
    bvh_root_callback: Callable[[BvhioJoint], None] | None = None,
    bvh_scaling_factor: float = 1.0,
    constrained_joint_ids: tuple[int, ...] = (0, 1, 2, 3, 4, 5, 6),
    neutral_qpos: np.ndarray | None = None,
    neutral_similarity_weight: float = 0.1,
    temporal_consistency_weight: float = 0.1,
    n_restarts: int = 3,
    error_acceptance_threshold: float = 1e-4,
    ftol: float = 1e-8,
    xtol: float = 1e-8,
    max_nfev: int = 2000,
    verbose: bool = False,
) -> MotionReferenceData:
    """Generates reference qpos and cartesian poses from BVH data.

    Args:
        model: The Mujoco model
        mj_base_id: The ID of the Mujoco base
        bvh_root: The root of the BVH tree
        bvh_to_mujoco_names: The mappings of BVH joints to Mujoco bodies
        bvh_base_id: The ID of the reference base (of the BVH file)
        ctrl_dt: The control timestep, used for time-based lookups.
        bvh_offset: Helps line up root with mj base
        bvh_root_callback: Modifies the root of the BVH tree (e.g. rotation)
        bvh_scaling_factor: The scaling factor for the reference motion
        constrained_joint_ids: The indices of the joints to constrain
        neutral_qpos: Helps with optimization, by default the starting qpos
        neutral_similarity_weight: Weight of neutral similarity term
        temporal_consistency_weight: Weight of temporal consistency term
        n_restarts: Number of random restarts to try
        error_acceptance_threshold: The threshold for the error
        ftol: The tolerance for the function value
        xtol: The tolerance for the solution
        max_nfev: The maximum number of function evaluations
        verbose: Whether to print verbose output

    Returns:
        A ReferenceMotionData object containing qpos and Cartesian poses.
    """
    if bvh_offset is None:
        bvh_offset = np.array([0.0, 0.0, 0.0])

    # 1. Generate Cartesian Poses
    np_cartesian_motion = get_reference_cartesian_poses(
        mappings=bvh_to_mujoco_names,
        model=model,
        root=bvh_root,
        reference_base_id=bvh_base_id,
        root_callback=bvh_root_callback,
        scaling_factor=bvh_scaling_factor,
        offset=bvh_offset,
    )
    total_frames = list(np_cartesian_motion.values())[0].shape[0]
    body_ids = list(np_cartesian_motion.keys())

    # 2. Solve IK for Qpos
    data = mujoco.MjData(model)
    constrained_jnt_mask = np.zeros(data.qpos.shape, dtype=bool)
    constrained_jnt_mask[np.array(constrained_joint_ids)] = True

    if neutral_qpos is None:
        neutral_qpos = np.copy(data.qpos)
    previous_qpos = np.copy(neutral_qpos)

    qpos_reference_motion = []
    for frame in range(total_frames):
        # Reload pose for consistency if callback modifies it in place
        bvh_root.loadPose(frame)
        if bvh_root_callback:
            bvh_root_callback(bvh_root)

        cartesian_pose_at_frame = {body_id: np_cartesian_motion[body_id][frame] for body_id in body_ids}
        qpos = solve_multi_body_ik(
            data=data,
            model=model,
            mj_base_id=mj_base_id,
            constrained_jnt_mask=constrained_jnt_mask,
            cartesian_pose=cartesian_pose_at_frame,
            neutral_qpos=neutral_qpos,
            prev_qpos=previous_qpos,
            neutral_similarity_weight=neutral_similarity_weight,
            temporal_consistency_weight=temporal_consistency_weight,
            n_restarts=n_restarts,
            error_acceptance_threshold=error_acceptance_threshold,
            ftol=ftol,
            xtol=xtol,
            max_nfev=max_nfev,
            verbose=verbose,
        )
        qpos_reference_motion.append(qpos)
        previous_qpos = qpos  # Update previous qpos for the next frame

    # 3. Compute qvel
    qvel_reference_motion = []
    for frame in range(total_frames - 1):
        qvel = (qpos_reference_motion[frame + 1] - qpos_reference_motion[frame]) / ctrl_dt
        qvel_reference_motion.append(qvel)

    jnp_reference_qpos = jnp.array(qpos_reference_motion)
    jnp_cartesian_motion = jax.tree.map(lambda arr: xax.HashableArray(arr), np_cartesian_motion)
    jnp_qvel = jnp.array(qvel_reference_motion)

    # 3. Create and return the data object
    return MotionReferenceData(
        qpos=xax.HashableArray(jnp_reference_qpos),
        qvel=xax.HashableArray(jnp_qvel),
        cartesian_poses=jnp_cartesian_motion,
        ctrl_dt=ctrl_dt,
    )


def vis_entry_point() -> None:
    parser = argparse.ArgumentParser(description="Visualize reference motion from a config file")
    parser.add_argument("motion", type=str, help="Path to the configuration file")
    parser.add_argument("--model", type=str, required=True, help="Path to the Mujoco model")
    parser.add_argument("--base_name", type=str, default="pelvis", help="Name of the Mujoco base")
    args = parser.parse_args()

    reference_data = MotionReferenceData.load(args.motion)
    model = mujoco.MjModel.from_xml_path(args.model)
    mj_base_id = get_body_id(model, args.base_name)

    np_cartesian_motion = jax.tree.map(lambda x: np.asarray(x.array), reference_data.cartesian_poses)

    visualize_reference_motion(
        model=model,
        reference_qpos=np.asarray(reference_data.qpos.array),
        cartesian_motion=np_cartesian_motion,
        mj_base_id=mj_base_id,
    )


@dataclass
class IKParams:
    neutral_similarity_weight: float = field(
        default=0.1,
        metadata={
            "help": "Weight of neutral similarity term",
        },
    )
    temporal_consistency_weight: float = field(
        default=0.1,
        metadata={
            "help": "Weight of temporal consistency term",
        },
    )
    n_restarts: int = field(
        default=5,
        metadata={
            "help": "Number of random restarts to try",
        },
    )
    error_acceptance_threshold: float = field(
        default=1e-4,
        metadata={
            "help": "The threshold for the error",
        },
    )
    ftol: float = field(
        default=1e-8,
        metadata={
            "help": "The tolerance for the function value",
        },
    )
    xtol: float = field(
        default=1e-8,
        metadata={
            "help": "The tolerance for the solution",
        },
    )
    max_nfev: int = field(
        default=2000,
        metadata={
            "help": "The maximum number of function evaluations",
        },
    )


@dataclass
class ReferenceMotionGeneratorConfig:
    model_path: str = field(
        default=MISSING,
        metadata={
            "help": "Path to the Mujoco model",
        },
    )
    bvh_path: str = field(
        default=MISSING,
        metadata={
            "help": "Path to the BVH file",
        },
    )
    output_path: str = field(
        default=MISSING,
        metadata={
            "help": "Path to the output file",
        },
    )
    mj_base_name: str = field(
        default=MISSING,
        metadata={
            "help": "Name of the Mujoco base",
        },
    )
    bvh_base_name: str = field(
        default=MISSING,
        metadata={
            "help": "Name of the BVH base",
        },
    )
    mappings: list[str] = field(
        default_factory=list,
        metadata={
            "help": "List of mappings in 'BvhJointName:MjBodyName' format",
        },
    )
    ctrl_dt: float = field(
        default=0.02,
        metadata={
            "help": "The control timestep",
        },
    )
    bvh_offset: list[float] = field(
        default_factory=lambda: [0, 0, 0],
        metadata={
            "help": "Offset of the BVH root",
        },
    )
    rotate_bvh_euler: list[float] = field(
        default_factory=lambda: [0, 0, 0],
        metadata={
            "help": "Euler angles to rotate the BVH root",
        },
    )
    bvh_scaling_factor: float = field(
        default=1.0,
        metadata={
            "help": "Scaling factor for the reference motion",
        },
    )
    constrained_joint_ids: list[int] = field(
        default_factory=lambda: [0, 1, 2, 3, 4, 5, 6],
        metadata={
            "help": "Indices of the constrained joints",
        },
    )
    ik_params: IKParams = field(
        default_factory=IKParams,
        metadata={
            "help": "IK parameters",
        },
    )
    verbose: bool = field(
        default=False,
        metadata={"help": "Whether to print verbose output"},
    )

    @classmethod
    def from_cli_args(cls, args: Sequence[str] | None = None) -> Self:
        """Parses configuration from command line arguments and optional config file."""
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument("-f", "--config", type=Path, default=None, help="Path to the configuration file.")
        parsed_args, remaining_args = parser.parse_known_args(args)
        config_path: Path | None = parsed_args.config

        cfg = cast(Self, OmegaConf.structured(cls))

        if config_path is not None:
            if config_path.is_file():
                logger.info("Loading configuration from: %s", config_path)
                try:
                    file_conf = OmegaConf.load(config_path)
                    cfg = cast(Self, OmegaConf.merge(cfg, file_conf))
                except Exception as e:
                    logger.error("Error loading config file '%s': %s", config_path, e)
                    sys.exit(1)
            else:
                logger.warning("Config file specified but not found: %s", config_path)

        try:
            cli_conf = OmegaConf.from_cli(remaining_args)
            cfg = cast(Self, OmegaConf.merge(cfg, cli_conf))
        except ConfigKeyError as e:
            logger.error(
                "Invalid command line config override: %s. Check if the key exists in the configuration structure.", e
            )
            sys.exit(1)
        except Exception as e:
            logger.error("Error parsing command line arguments: %s", e)
            sys.exit(1)

        try:
            OmegaConf.resolve(cast(DictConfig, cfg))
            assert isinstance(cfg.mappings, ListConfig), "Config error: 'mappings' must be a list of strings."
            mapping_msg = "Config error: All 'mappings' must be strings in 'BvhJointName:MjBodyName' format."
            assert all(isinstance(m, str) and ":" in m for m in cfg.mappings), mapping_msg
            bvh_offset_msg = "Config error: 'bvh_offset' must be a list of 3 floats."
            assert isinstance(cfg.bvh_offset, ListConfig) and len(cfg.bvh_offset) == 3, bvh_offset_msg
            rotate_bvh_euler_msg = "Config error: 'rotate_bvh_euler' must be a list of 3 floats."
            assert isinstance(cfg.rotate_bvh_euler, ListConfig) and len(cfg.rotate_bvh_euler) == 3, rotate_bvh_euler_msg
            constrained_joint_ids_msg = "Config error: 'constrained_joint_ids' must be a list of integers."
            assert isinstance(cfg.constrained_joint_ids, ListConfig), constrained_joint_ids_msg

        except MissingMandatoryValue as e:
            logger.error("Missing required configuration value: %s", e)
            logger.error("Please provide it via command line (e.g., model_path=...) or in a config file.")
            sys.exit(1)
        except AssertionError as e:
            logger.error(e)
            sys.exit(1)
        except Exception as e:
            logger.error("Configuration Error: %s", e)
            sys.exit(1)

        return cfg


def main() -> None:
    colorlogging.configure(level=logging.INFO)

    cfg = ReferenceMotionGeneratorConfig.from_cli_args(sys.argv[1:])

    logger.info("Final configuration:\n%s", OmegaConf.to_yaml(cfg))

    try:
        parsed_mappings_list = []
        for mapping_str in cfg.mappings:
            parts = mapping_str.split(":")
            if len(parts) != 2:
                raise ValueError("Invalid mapping format: '%s'. Use 'BvhJointName:MjBodyName'.", mapping_str)
            parsed_mappings_list.append(MotionReferenceMapping(reference_joint_name=parts[0], mj_body_name=parts[1]))
        parsed_mappings: tuple[MotionReferenceMapping, ...] = tuple(parsed_mappings_list)
    except ValueError as e:
        logger.error("Configuration Error: %s", e)
        sys.exit(1)

    logger.info("Loading MuJoCo model from %s", cfg.model_path)
    model = mujoco.MjModel.from_xml_path(cfg.model_path)

    logger.info("Loading BVH file from %s", cfg.bvh_path)
    bvh_root = bvhio.readAsHierarchy(cfg.bvh_path)

    constrained_joint_ids = tuple(cfg.constrained_joint_ids)

    mj_base_id = get_body_id(model, cfg.mj_base_name)
    bvh_base_id = get_reference_joint_id(bvh_root, cfg.bvh_base_name)

    logger.info("Generating reference motion...")
    output_dir = Path(cfg.output_path).parent
    output_dir.mkdir(parents=True, exist_ok=True)

    def rotation_callback(root: BvhioJoint) -> None:
        euler_rotation = np.array(cfg.rotate_bvh_euler)
        quat = R.from_euler("xyz", euler_rotation).as_quat(scalar_first=True)
        root.applyRotation(glm.quat(*quat), bake=True)

    reference_data = generate_reference_motion(
        model=model,
        mj_base_id=mj_base_id,
        bvh_root=bvh_root,
        bvh_to_mujoco_names=parsed_mappings,
        bvh_base_id=bvh_base_id,
        ctrl_dt=cfg.ctrl_dt,
        bvh_offset=np.array(cfg.bvh_offset),
        bvh_root_callback=rotation_callback,
        bvh_scaling_factor=cfg.bvh_scaling_factor,
        constrained_joint_ids=constrained_joint_ids,
        neutral_qpos=None,
        neutral_similarity_weight=cfg.ik_params.neutral_similarity_weight,
        temporal_consistency_weight=cfg.ik_params.temporal_consistency_weight,
        n_restarts=cfg.ik_params.n_restarts,
        error_acceptance_threshold=cfg.ik_params.error_acceptance_threshold,
        ftol=cfg.ik_params.ftol,
        xtol=cfg.ik_params.xtol,
        max_nfev=cfg.ik_params.max_nfev,
        verbose=cfg.verbose,
    )

    logger.info("Saving reference motion data to %s", cfg.output_path)
    reference_data.save(cfg.output_path)

    logger.info("Done.")


if __name__ == "__main__":
    main()



"""Tests for termination builders in the ksim package."""

from dataclasses import dataclass

import attrs
import jax
import jax.numpy as jnp
import mujoco
from jaxtyping import Array

import ksim


@attrs.define(frozen=True)
class DummyTermination(ksim.Termination):
    """Dummy termination for testing."""

    def __call__(self, state: ksim.PhysicsData, curriculum_level: Array) -> Array:
        return jnp.zeros((1,), dtype=jnp.bool_)


class DummyContact:
    """Mock contact data for testing."""

    geom1: Array = jnp.array([0, 2])
    geom2: Array = jnp.array([1, 3])
    dist: Array = jnp.array([-0.01, -0.02])


@jax.tree_util.register_dataclass
@dataclass
class DummyMjxData:
    """Mock mjx.Data for testing."""

    def __init__(self, has_contact: bool = True) -> None:
        self._has_contact = has_contact
        self._qpos = jnp.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0])

    @property
    def qpos(self) -> Array:
        return self._qpos

    @qpos.setter
    def qpos(self, value: Array) -> None:
        self._qpos = value

    @property
    def ncon(self) -> int:
        return 2 if self._has_contact else 0

    @property
    def contact(self) -> DummyContact:
        return DummyContact()


def test_termination_name() -> None:
    """Test that termination names are correctly generated."""
    term = DummyTermination()
    assert term.get_name() == "dummy_termination"
    assert term.termination_name == "dummy_termination"


class TestPitchTooGreatTermination:
    """Tests for the PitchTooGreatTermination class."""

    def test_pitch_too_great_termination(self) -> None:
        """Test that the PitchTooGreatTermination terminates when pitch exceeds the max."""
        data = DummyMjxData()
        curriculum_level = jnp.array(0.0)

        # With a small max_pitch, should terminate
        term = ksim.PitchTooGreatTermination(max_pitch=0.1)
        result = term(data, curriculum_level)
        assert result.item()

        # With a large max_pitch, should not terminate
        term = ksim.PitchTooGreatTermination(max_pitch=10.0)
        result = term(data, curriculum_level)
        assert not result.item()


class TestRollTooGreatTermination:
    """Tests for the RollTooGreatTermination class."""

    def test_roll_too_great_termination(self) -> None:
        """Test that the RollTooGreatTermination terminates when roll exceeds the max."""
        data = DummyMjxData()
        curriculum_level = jnp.array(0.0)

        # With a small max_roll, should terminate
        term = ksim.RollTooGreatTermination(max_roll=0.1)
        result = term(data, curriculum_level)
        assert result.item()

        # With a large max_roll, should not terminate
        term = ksim.RollTooGreatTermination(max_roll=10.0)
        result = term(data, curriculum_level)
        assert not result.item()


class TestMinimumHeightTermination:
    """Tests for the MinimumHeightTermination class."""

    def test_minimum_height_termination(self) -> None:
        """Test that the MinimumHeightTermination terminates when height is below the minimum."""
        # With min_height above the current height, should terminate
        # The default qpos[2] in DummyMjxData is 3.0
        data = DummyMjxData()
        curriculum_level = jnp.array(0.0)

        term = ksim.MinimumHeightTermination(min_height=4.0)
        result = term(data, curriculum_level)
        assert result.item()

        # With min_height below the current height, should not terminate
        term = ksim.MinimumHeightTermination(min_height=2.0)
        result = term(data, curriculum_level)
        assert not result.item()


class TestIllegalContactTermination:
    """Tests for the IllegalContactTermination class."""

    def test_illegal_contact_termination(self) -> None:
        """Test that the IllegalContactTermination terminates when there's an illegal contact."""
        # Test with illegal contact present (original DummyMjxData has geom1=[0,2] and geom2=[1,3])
        data = DummyMjxData()
        curriculum_level = jnp.array(0.0)

        term = ksim.IllegalContactTermination(illegal_geom_idxs=jnp.array([0, 3]))
        result = term(data, curriculum_level)
        assert result.item()

        # Test with no illegal contacts
        term = ksim.IllegalContactTermination(illegal_geom_idxs=jnp.array([4, 5]))
        result = term(data, curriculum_level)
        assert not result.item()

        # Test with no contacts at all
        data_no_contact = DummyMjxData(has_contact=False)
        term = ksim.IllegalContactTermination(illegal_geom_idxs=jnp.array([0, 1]))
        result = term(data_no_contact, curriculum_level)
        assert not result.item()

    def test_illegal_contact_termination_builder(self, humanoid_model: mujoco.MjModel) -> None:
        """Test that the IllegalContactTerminationBuilder creates a termination function."""
        geom_names = ["hand_left", "hand_right"]
        term = ksim.IllegalContactTermination.create(humanoid_model, geom_names=geom_names)
        assert term.termination_name == "illegal_contact_termination"
        assert term.illegal_geom_idxs.shape == (2,)



"""Walking default humanoid task with reference gait tracking."""

from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Generic, TypeVar

import attrs
import jax
import jax.numpy as jnp
import mujoco
import numpy as np
import xax

try:
    import bvhio
    import glm
    from bvhio.lib.hierarchy import Joint as BvhioJoint
except ImportError as e:
    raise ImportError(
        "In order to use reference motion utilities, please install Bvhio, using 'pip install bvhio'."
    ) from e


from jaxtyping import Array
from scipy.spatial.transform import Rotation as R

import ksim
from ksim.types import PhysicsModel
from ksim.utils.priors import (
    MotionReferenceMapping,
    generate_reference_motion,
    get_reference_joint_id,
    local_to_absolute,
    visualize_reference_motion,
    visualize_reference_points,
)

from .walking import HumanoidWalkingTask, HumanoidWalkingTaskConfig


@dataclass
class HumanoidWalkingReferenceMotionTaskConfig(HumanoidWalkingTaskConfig):
    bvh_path: str = xax.field(
        value=str(Path(__file__).parent / "data" / "walk-relaxed_actorcore.bvh"),
        help="The path to the BVH file.",
    )
    rotate_bvh_euler: tuple[float, float, float] = xax.field(
        value=(0, 0, 0),
        help="Optional rotation to ensure the BVH tree matches the Mujoco model.",
    )
    bvh_scaling_factor: float = xax.field(
        value=1.0,
        help="Scaling factor to ensure the BVH tree matches the Mujoco model.",
    )
    bvh_offset: tuple[float, float, float] = xax.field(
        value=(0.0, 0.0, 0.0),
        help="Offset to ensure the BVH tree matches the Mujoco model.",
    )
    mj_base_name: str = xax.field(
        value="pelvis",
        help="The Mujoco body name of the base of the humanoid",
    )
    reference_base_name: str = xax.field(
        value="CC_Base_Pelvis",
        help="The BVH joint name of the base of the humanoid",
    )
    visualize_reference_points: bool = xax.field(
        value=False,
        help="Whether to visualize the reference points.",
    )
    visualize_reference_motion: bool = xax.field(
        value=False,
        help="Whether to visualize the reference motion after running IK.",
    )


HUMANOID_REFERENCE_MAPPINGS = (
    MotionReferenceMapping("CC_Base_L_ThighTwist01", "thigh_left"),  # hip
    MotionReferenceMapping("CC_Base_L_CalfTwist01", "shin_left"),  # knee
    MotionReferenceMapping("CC_Base_L_Foot", "foot_left"),  # foot
    MotionReferenceMapping("CC_Base_L_UpperarmTwist01", "upper_arm_left"),  # shoulder
    MotionReferenceMapping("CC_Base_L_ForearmTwist01", "lower_arm_left"),  # elbow
    MotionReferenceMapping("CC_Base_L_Hand", "hand_left"),  # hand
    MotionReferenceMapping("CC_Base_R_ThighTwist01", "thigh_right"),  # hip
    MotionReferenceMapping("CC_Base_R_CalfTwist01", "shin_right"),  # knee
    MotionReferenceMapping("CC_Base_R_Foot", "foot_right"),  # foot
    MotionReferenceMapping("CC_Base_R_UpperarmTwist01", "upper_arm_right"),  # shoulder
    MotionReferenceMapping("CC_Base_R_ForearmTwist01", "lower_arm_right"),  # elbow
    MotionReferenceMapping("CC_Base_R_Hand", "hand_right"),  # hand
)


Config = TypeVar("Config", bound=HumanoidWalkingReferenceMotionTaskConfig)


def create_tracked_marker_update_fn(
    body_id: int, mj_base_id: int, tracked_pos_fn: Callable[[ksim.Trajectory], xax.FrozenDict[int, Array]]
) -> Callable[[ksim.Marker, ksim.Trajectory], None]:
    """Factory function to create a marker update for the tracked positions."""

    def _actual_update_fn(marker: ksim.Marker, transition: ksim.Trajectory) -> None:
        tracked_pos = tracked_pos_fn(transition)
        abs_pos = local_to_absolute(transition.xpos, tracked_pos[body_id], mj_base_id)
        marker.pos = tuple(abs_pos)

    return _actual_update_fn


def create_target_marker_update_fn(
    body_id: int, mj_base_id: int, target_pos_fn: Callable[[ksim.Trajectory], xax.FrozenDict[int, Array]]
) -> Callable[[ksim.Marker, ksim.Trajectory], None]:
    """Factory function to create a marker update for the target positions."""

    def _target_update_fn(marker: ksim.Marker, transition: ksim.Trajectory) -> None:
        target_pos = target_pos_fn(transition)
        abs_pos = local_to_absolute(transition.xpos, target_pos[body_id], mj_base_id)
        marker.pos = tuple(abs_pos)

    return _target_update_fn


@attrs.define(frozen=True, kw_only=True)
class QposReferenceMotionReward(ksim.Reward):
    reference_qpos: xax.HashableArray
    ctrl_dt: float
    norm: xax.NormType = attrs.field(default="l1")
    sensitivity: float = attrs.field(default=5.0)

    @property
    def num_frames(self) -> int:
        return self.reference_qpos.array.shape[0]

    def get_reward(self, trajectory: ksim.Trajectory) -> Array:
        qpos = trajectory.qpos
        step_number = jnp.int32(jnp.round(trajectory.timestep / self.ctrl_dt)) % self.num_frames
        reference_qpos = jnp.take(self.reference_qpos.array, step_number, axis=0)
        error = xax.get_norm(reference_qpos - qpos, self.norm)
        mean_error = error.mean(axis=-1)
        reward = jnp.exp(-mean_error * self.sensitivity)
        return reward


class HumanoidWalkingReferenceMotionTask(HumanoidWalkingTask[Config], Generic[Config]):
    reference_qpos: Array

    def get_rewards(self, physics_model: ksim.PhysicsModel) -> list[ksim.Reward]:
        rewards = [
            ksim.StayAliveReward(scale=1.0),
            QposReferenceMotionReward(
                reference_qpos=self.reference_motion.qpos, ctrl_dt=self.config.ctrl_dt, scale=0.5
            ),
            # ksim.NaiveForwardReward(scale=1.0),
        ]

        return rewards

    def run(self) -> None:
        mj_model: PhysicsModel = self.get_mujoco_model()
        root: BvhioJoint = bvhio.readAsHierarchy(self.config.bvh_path)
        reference_base_id = get_reference_joint_id(root, self.config.reference_base_name)
        mj_base_id = mujoco.mj_name2id(mj_model, mujoco.mjtObj.mjOBJ_BODY, self.config.mj_base_name)

        def rotation_callback(root: BvhioJoint) -> None:
            euler_rotation = np.array(self.config.rotate_bvh_euler)
            quat = R.from_euler("xyz", euler_rotation).as_quat(scalar_first=True)
            root.applyRotation(glm.quat(*quat), bake=True)

        reference_motion = generate_reference_motion(
            model=mj_model,
            mj_base_id=mj_base_id,
            bvh_root=root,
            bvh_to_mujoco_names=HUMANOID_REFERENCE_MAPPINGS,
            bvh_base_id=reference_base_id,
            bvh_offset=np.array(self.config.bvh_offset),
            bvh_root_callback=rotation_callback,
            bvh_scaling_factor=self.config.bvh_scaling_factor,
            ctrl_dt=self.config.ctrl_dt,
            neutral_qpos=None,
            neutral_similarity_weight=0.1,
            temporal_consistency_weight=0.1,
            n_restarts=3,
            error_acceptance_threshold=1e-4,
            ftol=1e-8,
            xtol=1e-8,
            max_nfev=2000,
            verbose=False,
        )
        self.reference_motion = reference_motion
        np_cartesian_motion = jax.tree.map(np.asarray, self.reference_motion.cartesian_poses)

        if self.config.visualize_reference_points:
            visualize_reference_points(
                model=mj_model,
                base_id=mj_base_id,
                reference_motion=np_cartesian_motion,
            )
        elif self.config.visualize_reference_motion:
            visualize_reference_motion(
                model=mj_model,
                reference_qpos=np.asarray(self.reference_motion.qpos),
                cartesian_motion=np_cartesian_motion,
                mj_base_id=mj_base_id,
            )
        else:
            super().run()


if __name__ == "__main__":
    # To run training, use the following command:
    #   python -m examples.walking_reference_motion
    # To visualize the environment, use the following command:
    #   python -m examples.walking_reference_motion run_model_viewer=True
    # On MacOS or other devices with less memory, you can change the number
    # of environments and batch size to reduce memory usage. Here's an example
    # from the command line:
    #   python -m examples.walking_reference_motion num_envs=2 batch_size=1
    HumanoidWalkingReferenceMotionTask.launch(
        HumanoidWalkingReferenceMotionTaskConfig(
            num_envs=2048,
            batch_size=256,
            num_passes=10,
            iterations=8,
            ls_iterations=8,
            epochs_per_log_step=1,
            valid_every_n_steps=10,
            # Simulation parameters.
            dt=0.002,
            ctrl_dt=0.02,
            max_action_latency=0.0,
            rollout_length_seconds=5.0,
            # PPO parameters
            gamma=0.97,
            lam=0.95,
            entropy_coef=0.001,
            learning_rate=3e-4,
            clip_param=0.3,
            max_grad_norm=1.0,
            # Gait matching parameters.
            bvh_path=str(Path(__file__).parent / "data" / "walk_normal_dh.bvh"),
            rotate_bvh_euler=(0, np.pi / 2, 0),
            bvh_scaling_factor=1 / 100,
            mj_base_name="pelvis",
            reference_base_name="CC_Base_Pelvis",
        ),
    )



"""Defines some Mujoco utility functions."""

__all__ = [
    "get_sensor_data_idxs_by_name",
    "get_qpos_data_idxs_by_name",
    "get_qvelacc_data_idxs_by_name",
    "get_ctrl_data_idx_by_name",
    "get_geom_data_idx_by_name",
    "get_geom_data_idx_from_name",
    "get_body_data_idx_by_name",
    "get_body_data_idx_from_name",
    "get_floor_idx",
    "geoms_colliding",
    "get_joint_metadata",
    "get_joint_names_in_order",
    "get_position_limits",
    "get_torque_limits",
    "update_model_field",
    "update_data_field",
    "slice_update",
    "quat_to_mat",
    "mat_to_quat",
    "get_body_pose",
    "get_geom_pose",
    "get_site_pose",
    "get_body_pose_by_name",
    "get_geom_pose_by_name",
    "get_site_pose_by_name",
    "remove_mujoco_joints_except",
    "add_new_mujoco_body",
]

import logging
from typing import Any, Hashable, TypeVar
from xml.etree import ElementTree as ET

import chex
import jax
import jax.numpy as jnp
import mujoco
import numpy as np
from jaxtyping import Array
from kscale.web.gen.api import JointMetadataOutput
from mujoco import mjx

from ksim.types import PhysicsData, PhysicsModel

logger = logging.getLogger(__name__)

Tk = TypeVar("Tk", bound=Hashable)
Tv = TypeVar("Tv")


def get_sensor_data_idxs_by_name(physics_model: PhysicsModel) -> dict[str, tuple[int, int | None]]:
    """Get mappings from sensor names to their data indices."""
    sensor_mappings = {}
    for i in range(len(physics_model.sensor_adr)):
        start = physics_model.sensor_adr[i]
        end = physics_model.sensor_adr[i + 1] if i < len(physics_model.sensor_adr) - 1 else None

        name_start = physics_model.name_sensoradr[i]
        name = bytes(physics_model.names[name_start:]).decode("utf-8").split("\x00")[0]
        sensor_mappings[name] = (start, end)
    return sensor_mappings


def get_qpos_data_idxs_by_name(physics_model: PhysicsModel) -> dict[str, tuple[int, int | None]]:
    """Get mappings from joint names to their position indices."""
    qpos_mappings = {}
    for i in range(len(physics_model.jnt_qposadr)):
        start = physics_model.jnt_qposadr[i]
        end = physics_model.jnt_qposadr[i + 1] if i < len(physics_model.jnt_qposadr) - 1 else None

        name_start = physics_model.name_jntadr[i]
        name = bytes(physics_model.names[name_start:]).decode("utf-8").split("\x00")[0]
        qpos_mappings[name] = (start, end)
    return qpos_mappings


def get_qvelacc_data_idxs_by_name(physics_model: PhysicsModel) -> dict[str, tuple[int, int | None]]:
    """Get mappings from joint names to their velocity/acceleration indices."""
    qvelacc_mappings = {}
    for i in range(len(physics_model.jnt_dofadr)):
        start = physics_model.jnt_dofadr[i]
        end = physics_model.jnt_dofadr[i + 1] if i < len(physics_model.jnt_dofadr) - 1 else None

        name_start = physics_model.name_jntadr[i]
        name = bytes(physics_model.names[name_start:]).decode("utf-8").split("\x00")[0]
        qvelacc_mappings[name] = (start, end)
    return qvelacc_mappings


def get_ctrl_data_idx_by_name(physics_model: PhysicsModel) -> dict[str, int]:
    """Get mappings from actuator names to their control indices."""
    ctrl_mappings = {}
    for i in range(len(physics_model.name_actuatoradr)):
        name_start = physics_model.name_actuatoradr[i]
        name = bytes(physics_model.names[name_start:]).decode("utf-8").split("\x00")[0]
        ctrl_mappings[name] = i
    return ctrl_mappings


def get_geom_data_idx_by_name(physics_model: PhysicsModel) -> dict[str, int]:
    """Get mappings from geometry names to their indices."""
    geom_mappings = {}
    for i in range(physics_model.ngeom):
        name_start = physics_model.name_geomadr[i]
        name = bytes(physics_model.names[name_start:]).decode("utf-8").split("\x00")[0]
        geom_mappings[name] = i
    return geom_mappings


def get_geom_data_idx_from_name(physics_model: PhysicsModel, geom_name: str) -> int:
    for i in range(physics_model.ngeom):
        name_start = physics_model.name_geomadr[i]
        name = bytes(physics_model.names[name_start:]).decode("utf-8").split("\x00")[0]
        if name == geom_name:
            return i
    choices = ", ".join(sorted(get_geom_data_idx_by_name(physics_model).keys()))
    raise KeyError(f"Geometry '{geom_name}' not found in model. Choices: {choices}")


def get_site_data_idx_by_name(physics_model: PhysicsModel) -> dict[str, int]:
    """Get mappings from site names to their indices."""
    site_mappings = {}
    for i in range(physics_model.nsite):
        name_start = physics_model.name_siteadr[i]
        name = bytes(physics_model.names[name_start:]).decode("utf-8").split("\x00")[0]
        site_mappings[name] = i
    return site_mappings


def get_site_data_idx_from_name(physics_model: PhysicsModel, site_name: str) -> int:
    """Get mappings from site names to their indices."""
    for i in range(physics_model.nsite):
        name_start = physics_model.name_siteadr[i]
        name = bytes(physics_model.names[name_start:]).decode("utf-8").split("\x00")[0]
        if name == site_name:
            return i
    choices = ", ".join(sorted(get_site_data_idx_by_name(physics_model).keys()))
    raise KeyError(f"Site '{site_name}' not found in model. Choices: {choices}")


def get_body_data_idx_by_name(physics_model: PhysicsModel) -> dict[str, int]:
    """Get mappings from body names to their indices."""
    body_mappings = {}
    for i in range(physics_model.nbody):
        name_start = physics_model.name_bodyadr[i]
        name = bytes(physics_model.names[name_start:]).decode("utf-8").split("\x00")[0]
        body_mappings[name] = i
    return body_mappings


def get_body_data_idx_from_name(physics_model: PhysicsModel, body_name: str) -> int:
    for i in range(physics_model.nbody):
        name_start = physics_model.name_bodyadr[i]
        name = bytes(physics_model.names[name_start:]).decode("utf-8").split("\x00")[0]
        if name == body_name:
            return i
    choices = ", ".join(sorted(get_body_data_idx_by_name(physics_model).keys()))
    raise KeyError(f"Body '{body_name}' not found in model. Choices: {choices}")


def get_floor_idx(physics_model: PhysicsModel, floor_name: str = "floor") -> int | None:
    """Get the index of the floor geometry."""
    geom_mappings = get_geom_data_idx_by_name(physics_model)
    assert floor_name in geom_mappings, f"Floor name {floor_name} not found in model"
    return geom_mappings[floor_name]


def geoms_colliding(state: PhysicsData, geom1: Array, geom2: Array) -> Array:
    """Return True if the geoms are colliding."""
    chex.assert_shape(geom1, (None,))
    chex.assert_shape(geom2, (None,))

    def get_colliding_inner(geom: Array, dist: Array, geom1: Array, geom2: Array) -> Array:
        # Create all pairs of geom indices
        g1, g2 = jnp.meshgrid(geom1, geom2, indexing="ij")

        def when_contacts_exist() -> Array:
            # geom shape is (n_contacts, 2)
            contact_geoms = geom
            contact_dists = dist  # shape (n_contacts,)

            g1_expanded = g1[..., jnp.newaxis]  # Shape: (n_geom1, n_geom2, 1)
            g2_expanded = g2[..., jnp.newaxis]  # Shape: (n_geom1, n_geom2, 1)

            # Expand contact_geoms to shape (1, 1, n_contacts, 2)
            contacts_expanded = contact_geoms[jnp.newaxis, jnp.newaxis, :]  # Shape: (1, 1, n_contacts, 2)

            # Check for matches in both orders (forward and reversed)
            forward_match = (g1_expanded == contacts_expanded[..., 0]) & (g2_expanded == contacts_expanded[..., 1])
            reverse_match = (g1_expanded == contacts_expanded[..., 1]) & (g2_expanded == contacts_expanded[..., 0])

            any_match = forward_match | reverse_match  # Shape: (n_geom1, n_geom2, n_contacts)

            negative_dist = contact_dists < 0  # Shape: (n_contacts,)
            collision_mask = (
                any_match & negative_dist[jnp.newaxis, jnp.newaxis, :]
            )  # Shape: (n_geom1, n_geom2, n_contacts)

            is_colliding = collision_mask.any(axis=2)  # Shape: (n_geom1, n_geom2)

            return is_colliding

        return jax.lax.cond(
            geom.shape[0] == 0,
            lambda _: jnp.zeros(g1.shape, dtype=jnp.bool_),
            lambda _: when_contacts_exist(),
            operand=None,
        )

    return get_colliding_inner(state.contact.geom, state.contact.dist, geom1, geom2)


def get_joint_names_in_order(model: PhysicsModel) -> list[str]:
    """Get the joint names in order of their indices."""
    return [bytes(model.names[model.name_jntadr[i] :]).decode("utf-8").split("\x00")[0] for i in range(model.njnt)]


def get_position_limits(model: PhysicsModel) -> dict[str, tuple[float, float]]:
    """Get the ranges of the joints."""
    ranges = {}
    for i in range(model.njnt):
        name_start = model.name_jntadr[i]
        name = bytes(model.names[name_start:]).decode("utf-8").split("\x00")[0]
        ranges[name] = (float(model.jnt_range[i, 0]), float(model.jnt_range[i, 1]))
    return ranges


def get_torque_limits(model: PhysicsModel) -> dict[str, tuple[float, float]]:
    """Get the torque limits of the joints."""
    ranges = {}
    for i in range(model.njnt):
        name_start = model.name_jntadr[i]
        name = bytes(model.names[name_start:]).decode("utf-8").split("\x00")[0]
        ranges[name] = (float(model.jnt_actfrcrange[i, 0]), float(model.jnt_actfrcrange[i, 1]))
    return ranges


def get_joint_metadata(
    model: PhysicsModel,
    kp: float | None = None,
    kd: float | None = None,
    armature: float | None = None,
    friction: float | None = None,
) -> dict[str, JointMetadataOutput]:
    """Get the joint metadata from the model."""
    metadata = {}
    for i in range(model.njnt):
        name_start = model.name_jntadr[i]
        name = bytes(model.names[name_start:]).decode("utf-8").split("\x00")[0]
        metadata[name] = JointMetadataOutput(
            kp=None if kp is None else str(kp),
            kd=None if kd is None else str(kd),
            armature=None if armature is None else str(armature),
            friction=None if friction is None else str(friction),
            id=None,
            flipped=None,
            offset=None,
            actuator_type=None,
        )
    return metadata


def update_model_field(model: mujoco.MjModel | mjx.Model, name: str, new_value: Array) -> mujoco.MjModel | mjx.Model:
    """Handles the update of a model field in a mujoco.MjModel or mjx.Model object.

    This is necessary because Mujoco handles model elements as pointers, while
    MJX model elements are stateful.
    """
    if isinstance(model, mujoco.MjModel):
        getattr(model, name)[:] = new_value
    else:
        model = model.replace(**{name: new_value})
    return model


def update_data_field(data: mujoco.MjData | mjx.Data, name: str, new_value: Array) -> mujoco.MjData | mjx.Data:
    """Handles the update of a data field in a mujoco.MjData or mjx.Data object.

    This is necessary because Mujoco handles data elements as pointers, while
    MJX data elements are stateful.

    Args:
        data: The mujoco.MjData or mjx.Data object to update.
        name: The name of the field to update.
        new_value: The new value to set the field to.

    Returns:
        The updated mujoco.MjData or mjx.Data object.
    """
    if isinstance(data, mujoco.MjData):
        getattr(data, name)[:] = new_value
    else:
        data = data.replace(**{name: new_value})
    return data


def slice_update(
    model: mujoco.MjModel | mjx.Model | mujoco.MjData | mjx.Data,
    name: str,
    slice: Any,  # noqa: ANN401
    value: Array,
) -> Array:
    """Update a slice of a model field."""
    if isinstance(model, (mujoco.MjModel, mujoco.MjData)):
        val = getattr(model, name).copy()
        val[slice] = value
        return val
    if isinstance(model, (mjx.Model, mjx.Data)):
        return getattr(model, name).at[slice].set(value)
    raise ValueError(f"Model type {type(model)} not supported")


def load_model(model: mujoco.MjModel) -> mjx.Model:
    mjx_model = mjx.put_model(model)
    mjx_model = jax.tree.map(jnp.array, mjx_model)
    return mjx_model


def quat_to_mat(quat: np.ndarray) -> np.ndarray:
    rot_mat = np.zeros(9, dtype=np.float64)
    mujoco.mju_quat2Mat(rot_mat, np.array(quat, dtype=np.float64))
    rot_mat = rot_mat.reshape(3, 3)
    return rot_mat


def mat_to_quat(mat: np.ndarray) -> np.ndarray:
    quat = np.zeros(4, dtype=np.float64)
    mujoco.mju_mat2Quat(quat, np.array(mat, np.float64).flatten())
    return quat


def get_body_pose(data: PhysicsData, body_idx: int) -> tuple[np.ndarray, np.ndarray]:
    position = data.xpos[body_idx].copy()
    quat = data.xquat[body_idx].copy()
    rot_mat = quat_to_mat(quat)
    return position, rot_mat


def get_geom_pose(data: PhysicsData, geom_idx: int) -> tuple[np.ndarray, np.ndarray]:
    position = data.geom_xpos[geom_idx].copy()
    rot_mat = data.geom_xmat[geom_idx].reshape(3, 3).copy()
    return position, rot_mat


def get_site_pose(data: PhysicsData, site_idx: int) -> tuple[np.ndarray, np.ndarray]:
    position = data.site_xpos[site_idx].copy()
    rot_mat = data.site_xmat[site_idx].reshape(3, 3).copy()
    return position, rot_mat


def get_body_pose_by_name(
    model: PhysicsModel,
    data: PhysicsData,
    body_name: str,
) -> tuple[np.ndarray, np.ndarray]:
    body_idx = get_body_data_idx_from_name(model, body_name)
    return get_body_pose(data, body_idx)


def get_geom_pose_by_name(
    model: PhysicsModel,
    data: PhysicsData,
    geom_name: str,
) -> tuple[np.ndarray, np.ndarray]:
    geom_idx = get_geom_data_idx_from_name(model, geom_name)
    return get_geom_pose(data, geom_idx)


def get_site_pose_by_name(
    model: PhysicsModel,
    data: PhysicsData,
    site_name: str,
) -> tuple[np.ndarray, np.ndarray]:
    site_idx = get_site_data_idx_from_name(model, site_name)
    return get_site_pose(data, site_idx)


def remove_mujoco_joints_except(file_path: str, joint_names: list[str]) -> str:
    """Remove all joints and references unless listed."""
    tree = ET.parse(file_path)
    root = tree.getroot()

    def dfs_remove_joints(element: ET.Element) -> None:
        for child in list(element):  # Use list to avoid modifying while iterating
            # Skip defaults as they are needed for armatures, etc.
            if child.tag == "default":
                continue

            if child.tag in {"joint", "freejoint"} and child.get("name") not in joint_names:
                element.remove(child)
            else:
                dfs_remove_joints(child)

    def dfs_remove_references(element: ET.Element) -> None:
        for child in list(element):
            # Check if the child references a joint not in joint_names
            joint_attr = child.get("joint")
            if joint_attr and joint_attr not in joint_names:
                element.remove(child)
            # Keyframes are difficult to reorder, so we remove them for now.
            elif child.tag == "keyframe":
                element.remove(child)
            else:
                dfs_remove_references(child)

    dfs_remove_joints(root)
    dfs_remove_references(root)

    # Re-write to file.
    return ET.tostring(root, encoding="utf-8").decode("utf-8")


def add_new_mujoco_body(
    file_path: str,
    parent_body_name: str,
    new_body_name: str,
    pos: tuple[float, float, float],
    quat: tuple[float, float, float, float],
    add_visual: bool = True,
    visual_geom_color: tuple[float, float, float, float] = (1, 0, 0, 1),
    visual_geom_size: tuple[float, float, float] = (0.05, 0.05, 0.05),
) -> str:
    """Add a new body to the model."""
    tree = ET.parse(file_path)
    root = tree.getroot()

    parent_body = None

    def dfs_find_body(element: ET.Element) -> ET.Element | None:
        for child in element:
            if child.tag == "body" and child.get("name") == parent_body_name:
                return child

            found = dfs_find_body(child)
            if found is not None:
                return found
        return None

    parent_body = dfs_find_body(root)
    if parent_body is None:
        raise ValueError(f"Parent body '{parent_body_name}' not found in model")

    # add the new body to the model
    new_body = ET.Element(
        "body",
        {
            "name": new_body_name,
            "pos": f"{pos[0]} {pos[1]} {pos[2]}",
            "quat": f"{quat[0]} {quat[1]} {quat[2]} {quat[3]}",
        },
    )

    if add_visual:
        visual_geom = ET.Element(
            "geom",
            {
                "name": f"{new_body_name}_visual",
                "type": "sphere",
                "class": "visual",
                "size": f"{visual_geom_size[0]} {visual_geom_size[1]} {visual_geom_size[2]}",
                "rgba": f"{visual_geom_color[0]} {visual_geom_color[1]} {visual_geom_color[2]} {visual_geom_color[3]}",
                "pos": "0 0 0",
                "quat": "1 0 0 0",
            },
        )
        new_body.append(visual_geom)

    parent_body.append(new_body)

    return ET.tostring(root, encoding="utf-8").decode("utf-8")
